"""Basic experiments for opioid simulator."""
import numpy as np

from whynot.framework import DynamicsExperiment, parameter
from whynot.simulators import opioid


def get_experiments():
    """Return all of the experiments for opioid simulator."""
    return [RCT, Confounding, UnobservedConfounding, Mediation]


def sample_initial_states(rng):
    """Sample initial states based on estimated uncertainty in Chen paper."""
    def sample(mean, std_dev):
        return mean + std_dev * rng.randn()

    nonmedical_users = sample(10029859, 100 * 329807)
    oud_users = sample(1369218, 100 * 116347)
    illicit_users = sample(328731, 100 * 60446)

    return opioid.State(nonmedical_users=nonmedical_users, oud_users=oud_users,
                        illicit_users=illicit_users)


@parameter(name="nonmedical_incidence_delta",
           default=-0.113, values=[-0.073, -0.113],
           description="Percent decrease in new nonmedical users of prescription opioids.")
def opioid_intervention(nonmedical_incidence_delta):
    """Return treatment config. Changes incidence of nonmedical abuse."""
    return opioid.Intervention(year=2015, nonmedical_incidence=nonmedical_incidence_delta)


def overdose_deaths(config, intervention, run):
    """Compute overdose deaths from the decade after intervention."""
    return opioid.simulator.compute_overdose_deaths(
        run, intervention.year, intervention.year + 10, config, intervention)


########################
# Randomized Experiment
#######################
#: RCT on the opioid simulator to see how reducing nonmedical opioid use reduces overdose deaths.
RCT = DynamicsExperiment(
    name="opioid_rct",
    description=("Randomized experiment reducing nonmedical incidence of "
                 "opioid use in 2015."),
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=0.5,
    outcome_extractor=overdose_deaths,
    covariate_builder=lambda run: run.initial_state.values())


########################
# Confounded experiments
########################
@parameter(name="propensity", default=0.9, values=[0.5, 0.6, 0.7, 0.9, 0.99],
           description="Probability of treatment assignment in positive group.")
def confounded_propensity_scores(config, intervention, control_runs, propensity):
    """Rollouts with top 10% of illicit deaths in 2015 more likely to receive treatment."""
    num_samples = len(control_runs)

    # More likely to assign treatment when illicit overdose
    # deaths are high in 2015
    illicit_deaths = [opioid.simulator.compute_illicit_deaths(run, 2015, config, intervention)
                      for run in control_runs]

    # Get top 10% of cities based on illicit overdose deaths
    sorted_idxs = sorted(range(num_samples), key=lambda idx: illicit_deaths[idx])
    top_death_idxs = sorted_idxs[int(num_samples * 0.9):]

    propensity_scores = (1. - propensity) * np.ones(num_samples)
    propensity_scores[top_death_idxs] = propensity

    return propensity_scores


# pylint: disable-msg=invalid-name
#: Observational version of RCT with confounding on observed nonmedical use overdose deaths.
Confounding = DynamicsExperiment(
    name="confounded_opioid_experiment",
    description="Opioid confounding by treating runs with high numbers of nonmedical use deaths.",
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=overdose_deaths,
    covariate_builder=lambda intervention, run: run[intervention.year].values())


# pylint: disable-msg=invalid-name
#: Observational version of RCT with unobserved confounding on nonmedical use overdose deaths.
UnobservedConfounding = DynamicsExperiment(
    name="unobserved_confounding_opioid_experiment",
    description=("Opioid unobserved confounding by treating runs with high "
                 "numbers of illicit deaths and omitting this variables."),
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=overdose_deaths,
    covariate_builder=lambda intervention, run: np.array([run[intervention.year].oud_users,
                                                          run[intervention.year].nonmedical_users]))


########################
# Mediation experiments
########################
@parameter(name="mediation_year", default=2020, values=[2017, 2019, 2020, 2021, 2024],
           description="Year of mediating variables--proxy for mediation strength.""")
def mediation_covariates(run, mediation_year):
    """Construct covariates for mediation experiment.

    Mediators are future state variables. Block backdoor paths with
    state variables in intervention year.
    """
    confounders = run[2015].values()
    mediation_state = run[mediation_year]
    mediators = np.array([mediation_state.nonmedical_users, mediation_state.oud_users])
    return np.concatenate([confounders, mediators])


# pylint: disable-msg=invalid-name
#: Observational version of RCT with mediation induced by observing future states.
Mediation = DynamicsExperiment(
    name="opioid_mediation",
    description="Mediation on opioid experiment by including future state.",
    simulator=opioid,
    simulator_config=opioid.Config(),
    intervention=opioid_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=confounded_propensity_scores,
    outcome_extractor=overdose_deaths,
    covariate_builder=mediation_covariates)

if __name__ == "__main__":
    RCT.run(num_samples=10)



#!/usr/bin/env python

from __future__ import print_function
from common import *
import subprocess
import string
import glob
from ctypes import windll

# From stackoverflow
def getAvailableDrives():
    drives = []
    bitmask = windll.kernel32.GetLogicalDrives()
    for letter in string.ascii_uppercase:
        if bitmask & 1:
            drives.append(letter)
        bitmask >>= 1

    return drives

def getPathHelper(exeName, possiblePaths):
    drives = getAvailableDrives()
    
    for d in drives:
        for p in possiblePaths:
            path = d + ":" + p
            fName = os.path.join(path, exeName)
            if os.path.exists(fName):
                return path

    return None

def getVCVarsPath():

    # Check if msbuild is already in the path
    try:
        subprocess.check_call(["msbuild", "/version"], stderr = DEVNULL, stdout = DEVNULL)
        return None # Nothing needs to done
    except:
        pass

    path = getPathHelper("vcvarsall.bat", [ r"\Program Files (x86)\Microsoft Visual Studio 14.0\VC" ])
    if not path:
        raise Exception("msbuild.exe is not in PATH, and cannot locate path of vcvarsall.bat (from VS2015).")
    return os.path.join(path, "vcvarsall.bat")

def getCMakePath():

    # Check if it's in the PATH
    try:
        subprocess.check_call(["cmake", "--version"], stderr = DEVNULL, stdout = DEVNULL)
        return None # Nothing needs to be added to the path
    except:
        pass

    # See if we can locate the executable
    path = getPathHelper("cmake.exe", [ r"\Program Files (x86)\CMake\bin" ])
    if not path:    
        raise Exception("cmake.exe is not in PATH, and cannot locate it.")
    return path

def getNSISPath():

    # Check if it's in the PATH
    try:
        subprocess.check_call(["makensis", "/VERSION"], stderr = DEVNULL, stdout = DEVNULL)
        return None # Nothing needs to be added to the path
    except:
        pass

    path = getPathHelper("makensis.exe", [ r"\Program Files (x86)\NSIS" ])
    if not path:    
        raise Exception("makensis.exe is not in PATH, and cannot locate it.")
    return path

def getGslAndTiffLibPrefix():

    def existsInPath(path, files):
        path = os.path.abspath(path)
        for n in files:
            if not os.path.exists(os.path.join(path, n)):
                return False

        return True


    paths = os.environ["PATH"].split(os.pathsep)
    for p in paths:
        if existsInPath(p, [ "gsl.dll", "cblas.dll", "tiff.dll" ]):
            upPath = os.path.abspath(os.path.join(p, ".."))
            libPath = os.path.join(upPath, "lib")
            incPath = os.path.join(upPath, "include")
            incGslPath = os.path.join(incPath, "gsl")
            if existsInPath(libPath, [ "gsl.lib", "cblas.lib", "tiff.lib" ]) and \
                    existsInPath(incPath, [ "tiff.h" ]) and \
                    existsInPath(incGslPath, [ "gsl_rng.h" ]):
                return upPath

    raise Exception("Unable to locate directory with GSL and TIFF libraries. Make sure that the files are installed and that the 'bin' direcory with dll files have been added to the PATH variable.")

def makeNSISFile(buildDir, srcDir, version, libDir):

    open(os.path.join(buildDir, "setpythonpath.bat"), "wt").write("""
setx PYTHONPATH "%PYTHONPATH%";%1
""")

    nsisStart = r"""
; The name of the installer
Name "SimpactCyan"

; The file to write
OutFile "SimpactCyan-{v}-install.exe"

; The default installation directory
InstallDir $PROGRAMFILES\SimpactCyan

; Request application privileges for Windows Vista
RequestExecutionLevel admin

;--------------------------------

; Pages

Page directory
Page instfiles
UninstPage uninstConfirm
UninstPage instfiles

;--------------------------------

; The stuff to install
Section "Installer section" ;No components page, name is not important

  ; Set output path to the installation directory.
  SetOutPath $INSTDIR
  
  ; Put file there
  File {b}\Release\simpact-cyan-release.exe
  File {b}\Debug\simpact-cyan-debug.exe
  File {b}\Release\maxart-release.exe
  File {b}\Debug\maxart-debug.exe
  File {l}\cblas.dll  
  File {l}\gsl.dll
  File {l}\tiff.dll
  WriteUninstaller uninstall.exe

  File {b}\setpythonpath.bat

  SetOutPath $INSTDIR\data
"""

    nsisEnd = r"""
  SetOutPath $INSTDIR\python
  File {s}\python\pysimpactcyan.py

  WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\SimpactCyan" "DisplayName" "SimpactCyan executables"
  WriteRegStr HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\SimpactCyan" "UninstallString" "$\"$INSTDIR\uninstall.exe$\""

  Exec '$INSTDIR\setpythonpath.bat "$INSTDIR\python"'
SectionEnd ; end the section

Section "un.Installer section"
  Delete $INSTDIR\setpythonpath.bat
  Delete $INSTDIR\simpact-cyan-release.exe
  Delete $INSTDIR\simpact-cyan-debug.exe
  Delete $INSTDIR\maxart-release.exe
  Delete $INSTDIR\maxart-debug.exe
  Delete $INSTDIR\cblas.dll  
  Delete $INSTDIR\gsl.dll
  Delete $INSTDIR\tiff.dll
  Delete $INSTDIR\uninstall.exe
  Delete $INSTDIR\python\pysimpactcyan.py
  RMDir /r $INSTDIR\data
  RMDir $INSTDIR\python
  RMDir $INSTDIR
  DeleteRegKey HKLM "Software\Microsoft\Windows\CurrentVersion\Uninstall\SimpactCyan"
SectionEnd
"""

    def getDataFiles(srcDir):
        startDir = os.getcwd()
        try:
            os.chdir(os.path.join(srcDir, "data"))
            return glob.glob("*") # TODO: for now there aren't any subdirs, so this might need to change later
        finally:
            os.chdir(startDir)

    dataFiles = getDataFiles(srcDir)
    nsisMiddle = ""
    for d in dataFiles:
        nsisMiddle += "  File {s}\\data\\" + d + "\n"

    nsiData = nsisStart + nsisMiddle + nsisEnd

    d = {
        "s": srcDir,
        "b": buildDir,
        "l": libDir,
        "v": version
    }

    nsiData = nsiData.format(**d)

    open(os.path.join(buildDir, "simpact.nsi"), "wt").write(nsiData)

def main():
    cmakePath = getCMakePath()
    print("CMake path is", cmakePath)

    vcvpath = getVCVarsPath()
    print("vcvarsall.bat path is", vcvpath)

    libsPrefix = getGslAndTiffLibPrefix()
    print("GSL and TIFF library prefix is", libsPrefix)

    nsisPath = getNSISPath()
    print("NSIS path is", nsisPath)

    path = sys.argv[1]
    buildDir, srcDir, version, workDir = createBuildAndSourceDir(path)

    makeNSISFile(buildDir, srcDir, version, os.path.join(libsPrefix, "bin"))

    batFile = os.path.join(buildDir, "buildit.bat")
    with open(batFile, "wt") as f:
        if cmakePath:
            f.write("set PATH=%PATH%;{}\n".format(cmakePath))
        if vcvpath:
            f.write('call "{}" x86\n'.format(vcvpath))
        if nsisPath:
            f.write("set PATH=%PATH%;{}\n".format(nsisPath))

        f.write('cmake -DCMAKE_INSTALL_PREFIX="{}" -G"Visual Studio 14 2015" -DCMAKE_CXX_FLAGS="-Dgsl_rng_get_default=gsl_rng_env_setup -DWIN32 /EHsc" "{}"\n'.format(libsPrefix, srcDir))
        f.write('if %errorlevel% neq 0 exit /b %errorlevel%\n')
        f.write('msbuild simpact-cyan.sln /p:Configuration=Release\n')
        f.write('if %errorlevel% neq 0 exit /b %errorlevel%\n')
        f.write('msbuild simpact-cyan.sln /p:Configuration=Debug\n')
        f.write('if %errorlevel% neq 0 exit /b %errorlevel%\n')
        f.write('makensis simpact.nsi\n')
        f.write('if %errorlevel% neq 0 exit /b %errorlevel%\n')
        
    os.chdir(buildDir)
    subprocess.check_call([batFile])

    print("")
    print("")
    print("")

    outputFile = os.path.join(buildDir, "SimpactCyan-{}-install.exe".format(version))
    if os.path.exists(outputFile):
        print("Output file is", outputFile)
    else:
        print("Build seems to have worked, or no error was detected, but output exe could not be located")

if __name__ == "__main__":
    main()



"""Unit tests for opioid epidemic simulator."""
import whynot as wn


def test_config():
    """Ensure intervention update works as expected."""
    intervention = wn.opioid.Intervention(year=2021, nonmedical_incidence=-.12)
    config = wn.opioid.Config()
    assert config.nonmedical_incidence.intervention_val == 0.
    config = config.update(intervention)
    assert config.nonmedical_incidence.intervention_val == -.12



"""Causal experiments for the Schelling model."""
import copy
import dataclasses

import numpy as np
import pandas as pd

from whynot.framework import GenericExperiment, parameter
from whynot.simulators import schelling
import whynot.utils as utils


def get_experiments():
    """Return all experiments for Schelling."""
    return [RCT]


################
# RCT Experiment
################
def run_simulator(control_config, treatment_config, seed):
    """Run the simulator in both the control and treatment conditions using a common seed."""
    control_segregation = schelling.simulate(control_config, rollouts=5, seed=seed)
    treatment_segregation = schelling.simulate(treatment_config, rollouts=5, seed=seed)
    return control_segregation, treatment_segregation


@parameter(name="education_pc", default=0.3, values=np.arange(0., 0.99, 0.05),
           description="Random fraction of the population to educate")
@parameter(name="education_boost", default=-1, values=[-1, -2, -3, -4],
           description="How much education decreases homophily")
def run_schelling(education_pc, education_boost, num_samples=100,
                  seed=None, show_progress=False, parallelize=True):
    """Run the basic RCT experiment on Schelling.

    Randomly assign Schelling grids to either treatment or control.
    Treatment corresponds to randomly educating some fraction of the
    grid to decrease their homophiliy and the outcome is the total
    fraction of segregated agents on the grids.

    Returns:
        - covariates
        - treatment
        - outcomes
        - ground_truth  (unit level treatment effects)

    """
    rng = np.random.RandomState(seed)

    configs = []
    for _ in range(num_samples):
        config = schelling.Config()
        config.height = 10
        config.width = 10
        config.homophily = 5
        config.density = rng.uniform(0.05, 0.6)
        config.minority_pc = rng.uniform(0.05, 0.45)
        config.education_boost = education_boost

        # By default, no education.
        config.education_pc = 0.0
        treatment_config = copy.deepcopy(config)
        # Treatment corresponds to education
        treatment_config.education_pc = education_pc

        seed = rng.randint(0, 99999)
        configs.append((config, treatment_config, seed))

    if parallelize:
        runs = utils.parallelize(run_simulator, configs, show_progress=show_progress)
    else:
        runs = [run_simulator(*args) for args in configs]

    control_outcomes = np.array([run[0] for run in runs])
    treatment_outcomes = np.array([run[1] for run in runs])

    # Potential confounders are density and minority pc
    dataframe = pd.DataFrame([dataclasses.asdict(config[0]) for config in configs])
    covariates = dataframe[["density", "minority_pc"]].values

    # First pass: Randomly assign treatment
    # We could only treat "cities" with high density and
    # high minority pc
    treatment = (rng.rand(num_samples) < 0.25).astype(np.int32)
    treatment_idxs = np.where(treatment == 1.0)[0]

    outcomes = np.copy(control_outcomes)
    outcomes[treatment_idxs] = treatment_outcomes[treatment_idxs]

    treatment_effects = treatment_outcomes - control_outcomes
    return (covariates, treatment, outcomes), treatment_effects


#: RCT experiment to understand the effect of homophily reduction on total segregation.
RCT = GenericExperiment(
    name="schelling_rct",
    description=("Test the efficacy of an education program to reduce "
                 "homophiliy of agents in the Schelling model."),
    run_method=run_schelling)



"""Simulator initialization."""

from whynot.simulators import civil_violence
from whynot.simulators import dice
from whynot.simulators import hiv
from whynot.simulators import lalonde
from whynot.simulators import lotka_volterra
from whynot.simulators import opioid
from whynot.simulators import schelling
from whynot.simulators import simpact_hiv
from whynot.simulators import world2
from whynot.simulators import world3



"""whynot package initialization."""

from whynot.algorithms import *
from whynot.simulators import *
from whynot import framework
from whynot.framework import Dataset, DynamicsExperiment, InferenceResult, parameter
from whynot import utils



"""Implementation of Adams et al.'s ODE simulator for HIV treatment.

Adams, Brian Michael, et al. Dynamic multidrug therapies for HIV: Optimal and
STI control approaches. North Carolina State University. Center for Research in
Scientific Computation, 2004. APA.

https://pdfs.semanticscholar.org/c030/127238b1dbad2263fba6b64b5dec7c3ffa20.pdf
"""
import dataclasses

from scipy.integrate import solve_ivp

import whynot as wn


@dataclasses.dataclass
class Config:
    # pylint: disable-msg=too-few-public-methods
    """Parameters for the simulation dynamics.

    Examples
    --------
    >>> # Run the simulation for 200 days with infected cell death rate 0.3
    >>> hiv.Config(duration=200, delta=0.3)

    """

    # pylint: disable-msg=invalid-name
    # pylint: disable-msg=too-many-instance-attributes
    #: Target cell type 1 production (source) rate
    lambda_1: float = 10000.
    #: Target cell type 2 production (source) rate
    lambda_2: float = 31.98
    #: Target cell type 1 death rate
    d_1: float = 0.01
    #: Target cell type 2 death rate
    d_2: float = 0.01
    #: Population 1 infection rate
    k_1: float = 8e-7
    #: Population 2 infection rate
    k_2: float = 1e-4
    #: Treatment efficacy reduction in population 2
    f: float = 0.34
    #: Infected cell death rate
    delta: float = 0.7
    #: Immune-induced clearance rate for population 1
    m_1: float = 1e-5
    #: Immune-induced clearance rate for population 2
    m_2: float = 1e-5
    #: Virions produced per infected cell
    N_T: float = 100
    #: Virus natural death rate
    c: float = 13
    #: Average number virions infecting a type 1 cell
    rho_1: float = 1
    #: Average number virions infecting a type 2 cell
    rho_2: float = 1
    #: Immune effector production (source) rate
    lambda_E: float = 1
    #: Maximum birth rate for immune effectors
    b_E: float = 0.3
    #: Saturation constant for immune effector birth
    K_B: float = 100
    #: Maximum death rate for immune effectors
    d_E: float = 0.3
    #: Saturation constant for immune effector death
    K_D: float = 500
    #: Natural death rate for immune effectors
    delta_E: float = 0.1

    # Treatment regimes-- only consider constant
    # treatments for now. The paper also references
    # a constant "fully efficacious" treatment with
    # e_1 = 0.7 and e_2 = 0.3.
    #: Drug efficacy
    epsilon_1: float = 0.0
    #: Efficacy of protease inhibitors
    epsilon_2: float = 0.0

    #: How many days to simulate
    duration: float = 400

    def update(self, intervention):
        """Generate a new config after applying the intervention."""
        return dataclasses.replace(self, **intervention.updates)


@dataclasses.dataclass
class State:
    # pylint: disable-msg=too-few-public-methods
    """State of the HIV simulator.

    The default state corresponds to the unstable "uninfected"
    steady state introduced by Adams et al. This state introduces
    one virus particle per ml of blood plasma and very low levels of infected
    T-cells.
    """

    # pylint: disable-msg=invalid-name
    #: Uninfected CD4+ T-lymphocytes (cells/ml)
    uninfected_T1: float = 1e6
    #: Infected CD4+ T-lymphocytes (cells/ml)
    infected_T1: float = 1e-4
    #: Uninfected macrophages (cells/ml)
    uninfected_T2: float = 3198
    #: Infected macrophages (cells/ml)
    infected_T2: float = 1e-4
    #: Free virus (copies/ml)
    free_virus: float = 1
    #: Immune response CTL E (cells/ml)
    immune_response: float = 10


class Intervention:
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the HIV model.

    Examples
    --------
    >>> # Starting in step 100, set epsilon_1 to 0.7 (leaving other variables unchanged)
    >>> Intervention(time=100, epsilon_1=0.7)

    """

    def __init__(self, time=100, **kwargs):
        """Specify an intervention in the dynamical system.

        Parameters
        ----------
            time: int
                Time of the intervention (days)
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        self.time = time
        config_args = set(f.name for f in dataclasses.fields(Config))
        for arg in kwargs:
            if arg not in config_args:
                raise TypeError("__init__() got an unexpected keyword argument {}!".format(arg))
        self.updates = kwargs


def virus_dynamics(time, state, config):
    """Update equations for the HIV simulaton.

    Currently, the update equations are time-invariant. However, the original
    paper considers different treatment policies (corresponding to time-varying
    versions of epsilon_1 and epsilon_2). This is a natural extension.
    Additionally, the dynamics do not currently implement a "shock" or a policy
    change of the type we study in other causal inference simulators.
    Once we add experiments, this is a natural extension.
    """
    # time is unused because the dynamics and config are time-invariant (for now).
    # pylint: disable-msg=unused-argument

    # Keep notation roughly consistent with the Adams paper.
    # pylint: disable-msg=invalid-name

    uninfected_T1, infected_T1, uninfected_T2, infected_T2, free_virus, immune_response = state

    delta_uninfected_T1 = (
        config.lambda_1
        - config.d_1 * uninfected_T1
        - (1 - config.epsilon_1) * config.k_1 * free_virus * uninfected_T1)

    delta_infected_T1 = (
        (1 - config.epsilon_1) * config.k_1 * free_virus * uninfected_T1
        - config.delta * infected_T1
        - config.m_1 * immune_response * infected_T1)

    delta_uninfected_T2 = (
        config.lambda_2
        - config.d_2 * uninfected_T2
        - (1 - config.f * config.epsilon_1) * config.k_2 * free_virus * uninfected_T2)

    delta_infected_T2 = (
        (1 - config.f * config.epsilon_1) * config.k_2 * free_virus * uninfected_T2
        - config.delta * infected_T2
        - config.m_2 * immune_response * infected_T2)

    delta_virus = (
        (1 - config.epsilon_2) * config.N_T * config.delta * (infected_T1 + infected_T2)
        - config.c * free_virus
        - free_virus
        * ((1. - config.epsilon_1) * config.rho_1 * config.k_1 * uninfected_T1
           + ((1. - config.f * config.epsilon_1) * config.rho_2 * config.k_2 * uninfected_T2)))

    delta_immune_response = (
        config.lambda_E
        + ((config.b_E * (infected_T1 + infected_T2))
           / (infected_T1 + infected_T2 + config.K_B)) * immune_response
        + ((config.d_E * (infected_T1 + infected_T2))
           / (infected_T1 + infected_T2 + config.K_D)) * immune_response
        - config.delta_E * immune_response)

    return [delta_uninfected_T1, delta_infected_T1, delta_uninfected_T2,
            delta_infected_T2, delta_virus, delta_immune_response]


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Adams HIV simulator model.

    The simulation starts at initial_state at time 0, and evolves the state
    using dynamics whose parameters are specified in config.

    Parameters
    ----------
        initial_state:  `whynot.simulators.hiv.State`
            Initial State object, which is used as x_{t_0} for the simulator.
        config:  `whynot.simulators.hiv.Config`
            Config object that encapsulates the parameters that define the dynamics.
        intervention: `whynot.simulators.hiv.Intervention`
            Intervention object that specifies what, if any, intervention to perform.
        seed: int
            Seed to set internal randomness. The simulator is deterministic, so
            the seed parameter is ignored.

    Returns
    -------
        run: `whynot.framework.Run`
            Rollout of the model.

    """
    # pylint: disable-msg=unused-argument

    def forward_map(time, state):
        """Wrap HIV dynamics update to pass in the config."""
        if intervention and time >= intervention.time:
            virus_dynamics(time, state, config.update(intervention))
        return virus_dynamics(time, state, config)

    # Numerically integrate the ODE
    solution = solve_ivp(forward_map, [0, config.duration], dataclasses.astuple(initial_state))

    times = solution.t
    states = [State(*state) for state in zip(*solution.y)]
    return wn.framework.Run(states=states, times=times)


if __name__ == "__main__":
    print(simulate(State(), Config()))



"""Interface to simpactcyan's agent-based HIV simulator."""
import os
import numpy as np
import pandas as pd
from whynot.simulators.simpact_hiv.simpactcyan.python import pysimpactcyan

# In contrast to the other simulators, this one currently doesn't
# take a namedtuple Config, since the configuration files for simpactcyan
# are sufficiently complicated and the API supports passing python dictionaries.


def compute_incidence(df_people, age_group, time_window):
    """Compute HVI incidence over the course of the experiment.

    Python modification from:
        https://github.com/wdelva/RSimpactHelp/blob/master/R/incidence.calculator.R

    Incidence is calculated in 3 steps.
          1. Calculate PY of exposure per person
          2. Calculate whether the person had the event or not
          3. Divide events by sum of PY.

    """
    # Calculate naive exposure time
    start_exposure = np.maximum(df_people.TOB + age_group[0], time_window[0])
    infect_time = df_people.InfectTime.astype(np.float64)
    end_exposure = np.minimum(
        np.minimum(df_people.TOB + age_group[1], time_window[1]),
        infect_time)

    exposure = end_exposure - start_exposure

    valid_incidents = (exposure > 0) &  \
        (infect_time > start_exposure) & (infect_time <= end_exposure)

    total_exposure = np.maximum(exposure, 0).sum()
    total_cases = valid_incidents.sum()

    return total_cases / total_exposure


def cumulative_incidence(results, config, age_group=(15, 50)):
    """Compute cumulative incidence total over year intervals for the given age group."""
    df_persons = pd.read_csv(results["logpersons"])

    incidences = []
    for year in list(range(config["population.simtime"]))[1:]:
        incidence = compute_incidence(df_persons, time_window=(year-1, year),
                                      age_group=age_group)
        incidences.append(incidence)

    return sum(incidences)


def simulate(config):
    """Run the simulator and return summary statistics."""
    simpact = pysimpactcyan.PySimpactCyan()

    # Manually set the locations of the binaries and datafiles
    # Trailing "/" is needed for data_path due to bug in Python
    # interface.
    current_dir = os.path.dirname(os.path.abspath(__file__))
    binary_path = os.path.join(current_dir, "simpactcyan_build")
    data_path = os.path.join(current_dir, "simpactcyan", "data") + "/"

    simpact.setSimpactDirectory(binary_path)
    simpact.setSimpactDataDirectory(data_path)

    results = simpact.run(config, "/tmp/simpacttest", parallel=True)
    return cumulative_incidence(results, config)



#!/usr/bin/env python

from __future__ import print_function
import pysimpactcyan
import oldpysimpactcyan
import sys
import random
import time

try:
    import comparesettings
except Exception as e:
    print("ERROR: {}".format(e))
    print("""
You need a 'comparesettings.py' file with settings for the output comparison
and speed comparison between different versions.

An example file could contain:

    # Config settings to compare the output files of different simpact cyan
    # executables
    outputTests = [ { "population.nummen": 100, "population.numwomen": 100 } ]

    # Config settings to compare the speed with which different simpact executables
    # run.
    speedTests = [ { "population.nummen": 1000, "population.numwomen": 1000 } ]

    # Each speed test will be run this amount of times, and the average and standard
    # deviation will be calculated
    speedTestRepeats = 10

""")
    sys.exit(-1)

def getSimpactForDir(d):

    for impl in [ pysimpactcyan, oldpysimpactcyan ]:

        s = impl.PySimpactCyan()
        s.setSimpactDirectory(d)
        try:
            s.setSimulationPrefix("simpact-cyan")
            print("Using {} for {}".format(impl.__name__, d))
            return s
        except:
            pass

    raise Exception("Neither the old nor the new simpact bindings seem to work with directory {}".format(d))

def calcAvgAndDev(tlist):
    if len(tlist) == 0:
        return (0.0,0.0)

    avg = 0.0
    for t in tlist:
        avg += t

    avg /= len(tlist)
    dev = 0.0
    for t in tlist:
        d = t - avg
        dev += d*d

    dev /= len(tlist)
    dev = dev**0.5

    return (avg, dev)

def main():

    if sys.version_info.major != 2 or sys.version_info.minor < 7:
        print("Need python 2.7")
        sys.exit(-1)

    try:
        dir1 = sys.argv[1]
        dir2 = sys.argv[2]
        outputDir = sys.argv[3]

        if len(sys.argv) > 4:
            raise Exception("Too many arguments")
    except Exception as e:
        print("Error: {}".format(e))
        print("Usage: {} dir1 dir2 outputdir".format(sys.argv[0]))
        sys.exit(-1)

    simpact1 = getSimpactForDir(dir1)
    simpact2 = getSimpactForDir(dir2)

    print("Ok, can use both directories")

    outputTests = comparesettings.outputTests

    print("TEST: RUNNING OUTPUT COMPARISONS")
    print("TEST: ==========================")
    for t in outputTests:
        for parallel in [ False, True ]:
            for opt in [ True, False ]:
                for release in [ True, False ]:
                    seed = int(random.random() * 1000000)
                    print("TEST: seed = {} parallel = {} opt = {} release = {}".format(seed,parallel,opt,release))
                    sys.stdout.flush()

                    if type(opt) == tuple:
                        opt1 = opt[0]
                        opt2 = opt[1]
                    else:
                        opt1 = opt
                        opt2 = opt

                    ret1 = simpact1.run(t, outputDir, parallel = parallel, opt = opt, release = release, seed = seed)
                    ret2 = simpact2.run(t, outputDir, parallel = parallel, opt = opt, release = release, seed = seed)

                    evt1 = open(ret1["logevents"]).read()
                    evt2 = open(ret2["logevents"]).read()
                    if evt1 == evt2:
                        print("TEST: PASS")
                    else:
                        print("TEST: FAILED")
                    sys.stdout.flush()

    print("TEST: RUNNING SPEED COMPARISONS")
    print("TEST: =========================")

    speedTests = comparesettings.speedTests
    repeats = comparesettings.speedTestRepeats

    for t in speedTests:
        parallel = False
        opt = True
        release = True
        seed = int(random.random() * 1000000)

        print("TEST: seed = {} parallel = {} opt = {} release = {}".format(seed,parallel,opt,release))
        sys.stdout.flush()

        times1 = [ ]
        times2 = [ ]
        for r in range(repeats):

            if type(opt) == tuple:
                opt1 = opt[0]
                opt2 = opt[1]
            else:
                opt1 = opt
                opt2 = opt

            t0 = time.time()
            ret1 = simpact1.run(t, outputDir, parallel = parallel, opt = opt, release = release, seed = seed)
            t1 = time.time()
            ret2 = simpact2.run(t, outputDir, parallel = parallel, opt = opt, release = release, seed = seed)
            t2 = time.time()

            dt1 = t1-t0
            dt2 = t2-t1

            times1.append(dt1)
            times2.append(dt2)

            evt1 = open(ret1["logevents"]).read()
            evt2 = open(ret2["logevents"]).read()
            if evt1 == evt2:
                print("TEST: PASS, DT1 = {}, DT2 = {}".format(dt1,dt2))
            else:
                print("TEST: FAILED")
            sys.stdout.flush()

        avg1, dev1 = calcAvgAndDev(times1)
        avg2, dev2 = calcAvgAndDev(times2)

        print("TEST: RUNTIME1 = {} +/- {}".format(avg1,dev1))
        print("TEST: RUNTIME2 = {} +/- {}".format(avg2,dev2))
        print("TEST: ------------------------------------------------------")

if __name__ == "__main__":
    main()



"""Unit tests for setup and intervention in world2 simulator."""
from whynot.simulators.world2.simulator import *


def test_intervention():
    """Simple test if intervention is correctly handled."""
    config = Config(birth_rate=0.2, pollution=0.1)
    intervention = Intervention(pollution=37, year=1960)
    initial_state = State()

    world = WorldDynamics(config, intervention, initial_state)

    assert world.config.birth_rate == 0.2 and world.config.pollution == 0.1
    world.step(time=1950, delta_t=0.2)
    assert world.config.birth_rate == 0.2 and world.config.pollution == 0.1
    world.step(time=1980, delta_t=0.2)
    assert world.config.birth_rate == 0.2 and world.config.pollution == 37
    world.step(time=2000, delta_t=0.2)
    assert world.config.birth_rate == 0.2 and world.config.pollution == 37
    world.step(time=1950, delta_t=0.2)
    assert world.config.birth_rate == 0.2 and world.config.pollution == 0.1



"""Default configurations for Lotka-Volterra simulation."""

DEFAULT_CONFIG = {
    "policy_year": 30,
    "growth_rate0": (1.5, 1.5),
    "carrying_capacity0": 100,
    "interaction01": (0.2, 0.2),
    "growth_rate1": (1.5, 1.5),
    "carrying_capacity1": 200,
    "interaction10": (1.6, 1.6),
}

DEFAULT_INITIAL_STATE = {
    "population0": 25,
    "population1": 80,
}

# Small change in the interaction coefficient
# after 30 years.
INTERVENTION_CONFIG = {
    "policy_year": 30,
    "growth_rate0": (1.5, 1.5),
    "carrying_capacity0": 100,
    "interaction01": (0.2, 0.2),
    "growth_rate1": (1.5, 1.5),
    "carrying_capacity1": 200,
    "interaction10": (1.6, 1.5),
}



#!/usr/bin/env python

import sys
import re
import os
import subprocess
import tempfile
import glob
import shutil
import pprint

def rm(n):
    try:
        os.unlink(n)
    except Exception as e:
        #print("Error deleting '{}': {}".format(n,e))
        pass

def testUsingVagrant(packageName, name, entry, workDir):
    print("Testing '{}'".format(name))

    vagrantFileData = open(entry["file"], "rt").read()
    vagrantFileData += """
Vagrant.configure(2) do |config|
  config.vm.provision "shell", inline: <<-SHELL
    cd /vagrant
    rm -f ERROR
    rm -f SUCCESS
    X=`which apt-get`
    if ! [ -z "$X" ] ; then
        sudo apt-get update -y
        sudo dpkg -i "{}"
        sudo apt-get install -f -y
    else
        X=`which dnf`
        if ! [ -z "$X" ] ; then
            dnf install -y ./{}
        else
            echo "System neither has apt-get not dnf" > ERROR
            exit -1
        fi
    fi

    # test the executable

    if ! simpact-cyan-release --showconfigoptions ; then
        echo "Unable to run simpact executable" > ERROR
        exit -1
    fi

    # test the python script
    source /etc/profile.d/simpactcyan.sh # adjust the python path

    PYTHON=python
    X=`which $PYTHON`
    if [ -z "$X" ] ; then
        PYTHON=python3
    fi

    if ! $PYTHON -c "import pysimpactcyan;simpact = pysimpactcyan.PySimpactCyan();r = simpact.run(None,'/tmp');print(r)" ; then
        echo "Unable to use python interface"
        exit -1
    fi

    touch SUCCESS
  SHELL
end
""".format(packageName, packageName)

    os.chdir(workDir)
    open("Vagrantfile", "wt").write(vagrantFileData)

    # Make sure that the status reporting files are gone, as well
    # as the old ssh config file
    rm("ERROR")
    rm("SUCCESS")
    rm("sshcfg")

    print("  Running vagrant.")
    with open("vagranttestlog-{}".format(name), "wb") as f:

        try:
            subprocess.check_call( [ "vagrant", "up" ], stdout=f, stderr=f)
            print("  Done.")

            if not os.path.exists("ERROR") and not os.path.exists("SUCCESS"):
                # It's possible that the directory was shared using rsync instead of sync,
                # try to copy the files

                # Get the ssh config
                with open("sshcfg","wb") as sshCfg:
                    subprocess.call(["vagrant", "ssh-config"], stdout=sshCfg, stderr=f)

                # And manually copy the files
                subprocess.call(["scp", "-F", "sshcfg", "default:/vagrant/ERROR", "default:/vagrant/SUCCESS", "."], stdout=f, stderr=f)

            if os.path.exists("ERROR"):
                raise Exception("Error during build")
            if not os.path.exists("SUCCESS"):
                raise Exception("No build error found, but success state not indicated")

            return True

        finally:
            subprocess.call( [ "vagrant", "destroy", "-f" ], stdout=f, stderr=f)

def buildUsingVagrant(fullFileName, name, entry, workDir):
    print("Building '{}'".format(name))

    if os.path.exists(fullFileName):
        fileName = os.path.basename(fullFileName)
        destFile = os.path.join(workDir, fileName)
        if not os.path.exists(destFile): # Don't copy it if not needed
            shutil.copy(fullFileName, destFile)
    else: # Assume it's an url
        fileName = fullFileName

    vagrantFileData = open(entry["file"], "rt").read()
    vagrantFileData += """
Vagrant.configure(2) do |config|
  config.vm.provision "shell", inline: <<-SHELL
    cd /vagrant
    rm -f ERROR
    rm -f SUCCESS
    X=`which python`
    if ! [ -z "$X" ] ; then
        CMD="python build.py {} --copypackage"
    else
        X=`which python3`
        if ! [ -z "$X" ] ; then
            CMD="python3 build.py {} --copypackage"
        else
            echo "Unable to find a python interpreter to run the build script" >> ERROR
            cat ERROR
            exit -1
        fi
    fi

    if ! $CMD ; then
        echo "Unable to build package" >> ERROR
    else
        touch SUCCESS
    fi
  SHELL
end
""".format(fileName, fileName)

    shutil.copy(entry["script"], os.path.join(workDir, "build.py"))
    shutil.copy("common.py", os.path.join(workDir, "common.py"))

    os.chdir(workDir)
    open("Vagrantfile", "wt").write(vagrantFileData)

    def getPackageFiles(useScp = False):
        if not useScp:
            return set(glob.glob("*.rpm") + glob.glob("*.deb"))

        out = subprocess.check_output(["ssh", "-F", "sshcfg", "default", "cd /vagrant ; echo *.deb ; echo *.rpm"])
        out = set(out.split())
        out.discard("*.deb")
        out.discard("*.rpm")
        return out
 
    # Make sure that the status reporting files are gone, as well
    # as the old ssh config file
    rm("ERROR")
    rm("SUCCESS")
    rm("sshcfg")

    print("  Running vagrant.")
    with open("vagrantbuildlog-{}".format(name), "wb") as f:

        try:
            beforeFiles = getPackageFiles()
            subprocess.check_call( [ "vagrant", "up" ], stdout=f, stderr=f)
            print("  Done.")

            useScp = False
            if not os.path.exists("ERROR") and not os.path.exists("SUCCESS"):
                # It's possible that the directory was shared using rsync instead of sync,
                # try to copy the files

                # Get the ssh config
                with open("sshcfg","wb") as sshCfg:
                    subprocess.call(["vagrant", "ssh-config"], stdout=sshCfg, stderr=f)

                # And manually copy the files
                subprocess.call(["scp", "-F", "sshcfg", "default:/vagrant/ERROR", "default:/vagrant/SUCCESS", "."], stdout=f, stderr=f)
                useScp = True

            if os.path.exists("ERROR"):
                raise Exception("Error during build")
            if not os.path.exists("SUCCESS"):
                raise Exception("No build error found, but success state not indicated")

            afterFiles = getPackageFiles(useScp)
            newFiles = afterFiles.difference(beforeFiles)

            #print(beforeFiles)
            #print(afterFiles)
            #print(newFiles)

            if len(newFiles) == 0:
                raise Exception("Vagrant build appeard to be successful, but no new files were found")
            if len(newFiles) > 1:
                raise Exception("Vagrant build appeard to be successful, more than one new file was found")

            outFile = [x for x in newFiles][0] # The new file

            if useScp and not os.path.exists(outFile):
                subprocess.check_call(["scp", "-F", "sshcfg", "default:/vagrant/" + outFile, "."], stdout=f, stderr=f)

            if not os.path.exists(outFile):
                raise Exception("Couldn't find the reported output file '{}'".format(outFile))

            return outFile
        finally:
            subprocess.call( [ "vagrant", "destroy", "-f" ], stdout=f, stderr=f)

cfgs = {   
    "12.04_32": { "file": "vagrantfiles/12.04_32/Vagrantfile", "script": "build-debian.py" },
    "12.04_64": { "file": "vagrantfiles/12.04_64/Vagrantfile", "script": "build-debian.py" },
    "14.04_32": { "file": "vagrantfiles/14.04_32/Vagrantfile", "script": "build-debian.py" },
    "14.04_64": { "file": "vagrantfiles/14.04_64/Vagrantfile", "script": "build-debian.py" },
    "16.04_32": { "file": "vagrantfiles/16.04_32/Vagrantfile", "script": "build-debian.py" },
    "16.04_64": { "file": "vagrantfiles/16.04_64/Vagrantfile", "script": "build-debian.py" },
    "fc23_64": { "file": "vagrantfiles/fc23_64/Vagrantfile", "script": "build-fedora.py" },
    "fc24_64": { "file": "vagrantfiles/fc24_64/Vagrantfile", "script": "build-fedora.py" },
    "debian8_64": { "file": "vagrantfiles/debian8_64/Vagrantfile", "script": "build-debian.py" },
}

def main():

    startDir = os.getcwd()



    def printConfigNames(cfgs):
        names = [ n for n in cfgs ]
        names.sort()
        for n in names:
            print("  {}".format(n))
        print("")
        return names

    try:
        if len(sys.argv) < 3:
            raise Exception("At least a source path and one configuration needs to be specified. Nothing to do now.")
        
        fullFileName = sys.argv[1]

        buildNames = set()
        for o in sys.argv[2:]:
            if o in cfgs: # Add it if it matches exactly
                buildNames.add(o)
            else: # Try to interpret it as a regular expression
                for cfgName in cfgs:
                    match = re.match(o, cfgName)
                    if not match or match.group(0) != cfgName:
                        pass
                    else:
                        buildNames.add(cfgName)

        if len(buildNames) == 0:
            raise Exception("No matching build configurations were detected")

    except Exception as e:
        print("""
Error: {}

Usage: {} file.tar.gz name1 [ name2 ... ]

Each name can be an exact configuration name, or can be a regular 
expression that matches one or more of these names.
""".format(e, sys.argv[0]))
        print("Valid configuration names are: ")
        printConfigNames(cfgs)
        sys.exit(-1)

    try:
        print("Checking if 'vagrant' exists")
        subprocess.check_call(["vagrant", "-v"])
    except subprocess.CalledProcessError:
        pass
    except Exception as e:
        print("""Error: {}

Couldn't run 'vagrant', which is needed to create the package files.
See https://www.vagrantup.com/
""".format(e))
        sys.exit(-1)
    
    workDir = tempfile.mkdtemp(prefix = "simpactbuildtmp")
    print("Work directory is " + workDir)

    print("Building the following configurations:")
    buildNames = printConfigNames(buildNames)

    results = [ ]
    for n in buildNames:
        entry = cfgs[n]

        cfgInfo = { "name": n,
                    "package": None,
                    "builderror": None,
                    "testsuccess": False,
                    "testerror": None }

        packageName = None
        try:
            os.chdir(startDir) # Make sure we always cleanly start in this directory

            packageName = buildUsingVagrant(fullFileName, n, entry, workDir)
            cfgInfo["package"] = packageName
            print("  Build ok: {}".format(packageName))
        except Exception as e:
            cfgInfo["builderror"] = str(e)
            print("  Build error: {}".format(e))

        if packageName:
            try:
                os.chdir(startDir) # Make sure we always cleanly start in this directory

                testUsingVagrant(packageName, n, entry, workDir)
                cfgInfo["testsuccess"] = True
                print("  Package test ok")
            except Exception as e:
                cfgInfo["testerror"] = str(e)
                print("  Package test error: {}".format(e))

        results.append(cfgInfo)

    #pprint.pprint(results)
    print("")
    print("Summary:")
    print("--------")
    for r in results:
        buildMsg = "undefined"
        testMsg = ""
        if r["package"]:
            buildMsg = r["package"]
            if r["testsuccess"]:
                testMsg = "OK"
            else:
                if r["testerror"]:
                    testMsg = r["testerror"]
                else:
                    testMsg = "unknown test error"

        else:
            if r["builderror"]:
                buildMsg = r["builderror"]
            else:
                buildMsg = "unknown build error"

        print("{}\t{}\t{}".format(r["name"], buildMsg, testMsg))

    print("")
    print("Work directory is " + workDir)

def testMain():
    workDir = "/tmp/simpactbuildtmp5m9fIl"
    n = "fc24_64"
    packageName = "simpactcyan-0.19.6-1.fc24.x86_64.rpm"

    testUsingVagrant(packageName, n, cfgs[n], workDir)

if __name__ == "__main__":
    #testMain()
    main()




"""Implementation of Chen et al. systems dynamic model for opioid abuse.

The model implementation and parameterization is taken from:
    Chen Q, Larochelle MR, Weaver DT, et al. Prevention of Prescription Opioid
    Misuse and Projected Overdose Deaths in the United States. JAMA Network
    Open. 2019;2(2):e187621. doi:10.1001/jamanetworkopen.2018.7621
"""
import copy
import dataclasses

import numpy as np
from scipy.integrate import solve_ivp

import whynot as wn


class TimeVaryingParam():  # pylint: disable=too-few-public-methods
    """Encapsulate a time-varying parameter obtained from a joinpoint model."""

    def __init__(self, baseline, alpha_1, joinpoint=None, alpha_2=None, intervention_year=None,
                 intervention_val=None, start_year=2002, stabilize_year=None):
        """Specify a joinpoint model for a time-varying parameter.

        Parameters
        ----------
            baseline: float
                Baseline value at initial time point
            alpha_1: float
                Annual percentage change (APC) before joinpoint
            joinpoint: float
                (Optional) time indicating signifcant change in APC
            alpha_2: float
                (Optional) Annual percentage change (APC) after joinpoint
            intervention_year: float
                (Optional) Year to intervene on the APC
            intervention_val: float
                (Optional) New value for the APC after intevention
            start_year: float
                (Optional) initial time point
            stabilize_year: float
                (Optional) Year (after 2015) to set APC decrease to 0.

        """
        if joinpoint is None and alpha_2 is not None:
            raise ValueError("Must specify joinpoint to use alpha_2")

        if stabilize_year is not None and stabilize_year <= 2015:
            raise ValueError("stabilize_year must be after 2015")

        if intervention_year is not None:
            if stabilize_year is not None:
                raise ValueError("Cannot use stabilization and intervention simultaneously")
            if joinpoint is not None and intervention_year < joinpoint:
                raise ValueError("Cannot intervene before joinpoint")

        self.baseline = baseline
        self.start_year = start_year
        self.alpha_1 = alpha_1
        self.joinpoint = joinpoint
        self.alpha_2 = alpha_2
        self.stabilize_year = stabilize_year
        self.intervention_year = intervention_year
        self.intervention_val = intervention_val

    def _unstabilized_value(self, time):
        """Return value of parameter without taking stabilization into account."""
        running_value = self.baseline
        if self.joinpoint is None or time <= self.joinpoint:
            return running_value * np.power(1. + self.alpha_1, time - self.start_year)

        running_value *= np.power(1. + self.alpha_1, self.joinpoint - self.start_year)

        if self.intervention_year is None or time <= self.intervention_year:
            return running_value * np.power(1. + self.alpha_2, time - self.joinpoint)

        running_value *= np.power(1. + self.alpha_2, self.intervention_year - self.joinpoint)

        return running_value * np.power(1. + self.intervention_val, time - self.intervention_year)

    def _current_rate(self, time):
        """Return current APC."""
        if self.joinpoint is None or time <= self.joinpoint:
            return self.alpha_1
        if self.intervention_year is None or time <= self.intervention_year:
            return self.alpha_2
        return self.intervention_val

    def __getitem__(self, time):
        """Retrieve the parameter value at time t."""
        if self.stabilize_year is None or time < 2015:
            return self._unstabilized_value(time)

        running_val = self._unstabilized_value(2015)
        rate = self._current_rate(time)

        # Decrease at a constant rate until stabilize year
        decay_per_year = rate / (self.stabilize_year - 2015)
        for year in range(2016, self.stabilize_year + 1):
            rate -= decay_per_year
            if time <= year:
                break
            running_val *= (1. + rate)

        # Handle fractional values
        if time <= self.stabilize_year:
            remainder = time - np.floor(time)
            running_val *= (1. + rate) ** remainder

        return running_val


@dataclasses.dataclass
class Config:
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of the dynamics for opioid epidemic simulator.

    By default, the confiruation uses the mean parameter values specified in
    eTable 4 and the pessimistic-case scenario (overdose mortality and incidence
    for illicit use stabilizes in 2025) and considers the scenario with no
    improvement in incidence after 2015.

    Examples
    --------
    >>> # Standard run of the dynamics with slightly rate of overdose
    >>> # deaths from nonmedial opioids.
    >>> opioid.Config(start_time=2002, end_time=2030, nonmedical_overdose=0.002)

    """

    # Dynamics parameters
    #: Annual incidence of nonmedical prescription opioid use.
    nonmedical_incidence: TimeVaryingParam = TimeVaryingParam(
        baseline=2496835, alpha_1=0.0016, joinpoint=2008,
        alpha_2=-0.0747, intervention_year=2015, intervention_val=0.0)

    #: Annual incidence of illicit opioid use from sources other than prescription opioids.
    illicit_incidence: TimeVaryingParam = TimeVaryingParam(
        baseline=13489, alpha_1=0.235, stabilize_year=2025)

    #: Annual transition rate of non-medical opioid use to opioid use disorder.
    nonmedical_to_oud: float = 0.06

    #: Annual transition rate of users with opioid user disorder to illicit opioid users.
    oud_to_illicit: TimeVaryingParam = TimeVaryingParam(baseline=0.071, alpha_1=0.042)

    #: Annual transition rate of nonmedial to illicit opioid users
    nonmedical_to_illicit: float = 0.002

    #: Overdose mortality rate for nonmedial prescription opioid use.
    nonmedical_overdose: float = 0.001

    # Decrease baseline slightly from paper to match the reported
    # calibration targets.
    #: Overdose mortality rate for opioid use disorder.
    oud_overdose: TimeVaryingParam = TimeVaryingParam(
        baseline=0.003, alpha_1=0.224, joinpoint=2007, alpha_2=0.03)
    #: Overdose mortality rate for illicit opioid use.
    illicit_overdose: TimeVaryingParam = TimeVaryingParam(
        baseline=0.005, alpha_1=0.011, joinpoint=2011, alpha_2=0.356, stabilize_year=2025)
    #: "Exit rate" from nonmedical users group. (Either stop using or die from non-opioid causes).
    nonmedical_exit: float = 0.177
    #: "Exit rate" from opioid user disorder group.
    oud_exit: TimeVaryingParam = TimeVaryingParam(baseline=0.311, alpha_1=-0.021)
    #: "Exit rate" of illicit opioid users.
    illicit_exit: float = 0.319

    # Simulator configuration
    #: Year to start the simulation
    start_time: float = 2002
    #: Year to end the simulation
    end_time: float = 2030

    def update(self, intervention):
        """Return a new config after applying intervention."""
        config = copy.deepcopy(self)
        if intervention.nonmedical_incidence is not None:
            config.nonmedical_incidence.intervention_year = intervention.year
            config.nonmedical_incidence.intervention_val = intervention.nonmedical_incidence
        if intervention.illicit_incidence is not None:
            config.illicit_incidence.intervention_year = intervention.year
            config.illicit_incidence.intervention_val = intervention.illicit_incidence
        return config


@dataclasses.dataclass
class Intervention:
    """Intervention parameters for opioid epidemic simulator.

    Currently, the simulator only supports interventions to reduce
    nonmedical and illicit opioid use because these are the only interventions
    considered in the originl Chen et al. paper.

    Examples
    --------
    >>> # Intervene in 2017 to reduce nonmedical incidence by 5%
    >>> opioid.Intervention(nonmedical_incidence=-0.05, year=2017)

    """

    # pylint: disable-msg=too-few-public-methods
    #: Year of intervention
    year: float = 2015

    #: Percent change in nonmedial opioid use (range -1. to 1.)
    nonmedical_incidence: float = None
    #: Percent change in illicit opioid use (range -1. to 1.)
    illicit_incidence: float = None


@dataclasses.dataclass
class State:
    # pylint: disable-msg=too-few-public-methods
    """State of the opioid simulator.

    Default values correspond to the mean initial state used in the Chen paper.
    """

    #: Non-medical users of prescription opioids
    nonmedical_users: float = 10029859
    #: Non-medical users of prescription opioid with prescription opioid use disorder.
    oud_users: float = 1369218
    #: Illicit opioid users (e.g. users of heroin/fentanyl as opposed to presciption opioids)
    illicit_users: float = 328371

    def values(self):
        """Return the state as a numpy array."""
        state_dict = dataclasses.asdict(self)
        return np.array([state_dict[key] for key in sorted(state_dict.keys())])


def compute_overdose_deaths(run, start_year, end_year, config, intervention):
    """Compute number of opioid overdose deaths from start_year to end_year."""
    if intervention:
        config = config.update(intervention)

    deaths = 0.
    for outcome_year in range(start_year, end_year + 1):
        state = run[outcome_year]
        nonmedical_deaths = state.nonmedical_users * config.nonmedical_overdose
        oud_deaths = state.oud_users * config.oud_overdose[outcome_year]
        illicit_deaths = state.illicit_users * config.illicit_overdose[outcome_year]
        deaths += nonmedical_deaths + oud_deaths + illicit_deaths
    return deaths


def compute_illicit_deaths(run, outcome_year, config, intervention):
    """Compute number of non-medical use overdose deaths in year."""
    if intervention:
        config = config.update(intervention)

    return run[outcome_year].illicit_users * config.illicit_overdose[outcome_year]


def dynamics(time, state, config):
    """State dynamics for the Chen et al. opioid epidemic model."""
    nonmedical_users, oud_users, illicit_users = state

    # Update non-medical opioid use cases
    nonmedical_increase = config.nonmedical_incidence[time]
    nonmedical_decrease = nonmedical_users * (
        config.nonmedical_exit + config.nonmedical_overdose +
        config.nonmedical_to_oud + config.nonmedical_to_illicit)

    nonmedical_delta = nonmedical_increase - nonmedical_decrease

    # Update opioid use disorder cases
    oud_increase = config.nonmedical_to_oud * nonmedical_users
    oud_decrease = oud_users * (
        config.oud_exit[time] + config.oud_overdose[time] + config.oud_to_illicit[time])
    oud_delta = oud_increase - oud_decrease

    # Update illicit opiate use cases
    illicit_increase = config.illicit_incidence[time] + (
        config.oud_to_illicit[time] * oud_users +
        config.nonmedical_to_illicit * nonmedical_users)

    illicit_decrease = illicit_users * (
        config.illicit_exit + config.illicit_overdose[time])

    illicit_delta = illicit_increase - illicit_decrease

    return [nonmedical_delta, oud_delta, illicit_delta]


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Chen et al. model from the given state and config.

    The simulator is currently deterministic, so the seed parameter is ignored.
    """
    # pylint: disable-msg=unused-argument
    if intervention:
        config = config.update(intervention)

    def forward_map(time, state):
        """Wrap dynamics to use problem config."""
        return dynamics(time, state, config)

    # Numerically integrate the ODE
    solution = solve_ivp(forward_map, [config.start_time, config.end_time],
                         dataclasses.astuple(initial_state),
                         t_eval=list(range(config.start_time, config.end_time + 1)))

    times = solution.t
    states = [State(*s) for s in zip(*solution.y)]
    return wn.framework.Run(states=states, times=times)



"""Simulation code for Lotka-Volterra model."""
import dataclasses

import numpy as np
from scipy.integrate import solve_ivp

import whynot as wn


@dataclasses.dataclass
class Config:
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of Lotka-Volterra dynamics.

    Examples
    --------
    >>> # Configure the simulator so each caught rabbit creates 2 foxes
    >>> lotka_volterra.Config(fox_growth=0.5)

    """

    # Dynamics parameters
    #: Natural growth rate of rabbits, when there's no foxes.
    rabbit_growth: float = 1.
    #: Natural death rate of rabbits, due to predation.
    rabbit_death: float = 0.1
    #: Natural death rate of fox, when there's no rabbits.
    fox_death: float = 1.5
    #: Factor describing how many caught rabbits create a new fox.
    fox_growth: float = 0.75

    # Simulator book-keeping
    #: Start time of the simulator.
    start_time: float = 0
    #: End time of the simulator.
    end_time: float = 100

    def update(self, intervention):
        """Generate a new config by applying the intervention."""
        return dataclasses.replace(self, **intervention.updates)


@dataclasses.dataclass
class State:
    # pylint: disable-msg=too-few-public-methods
    """State of the Lotka-Volterra model."""

    #: Number of rabbits.
    rabbits: float = 10.
    #: Number of foxes.
    foxes: float = 5.

    def values(self):
        """Return the state as a numpy array."""
        state_dict = dataclasses.asdict(self)
        return np.array([state_dict[key] for key in sorted(state_dict.keys())])


class Intervention:
    # pylint: disable-msg=too-few-public-methods
    """Parameterization of an intervention in the Lotka-Volterra model.

    An intervention changes a subset of the configuration variables in the
    specified year. The remaining variables are unchanged.

    Examples
    --------
    >>> # Starting in year 25, set fox_growth to 0.4 (leave other variables unchanged)
    >>> Intervention(year=25, fox_growth=0.4)

    """

    def __init__(self, year=30, **kwargs):
        """Specify an intervention in lotka_volterra.

        Parameters
        ----------
            year: int
                Time of intervention in simulator dynamics.
            kwargs: dict
                Only valid keyword arguments are parameters of Config.

        """
        self.year = year
        config_args = set(f.name for f in dataclasses.fields(Config))
        for arg in kwargs:
            if arg not in config_args:
                raise TypeError("__init__() got an unexpected keyword argument {}!".format(arg))
        self.updates = kwargs


def dynamics(state, config):
    """Noisy Lotka-Volterra equations for 2 species.

    Equations from:
        https://scipy-cookbook.readthedocs.io/items/LoktaVolterraTutorial.html
    """
    rabbits, foxes = state

    delta_rabbits = (config.rabbit_growth * rabbits
                     - config.rabbit_death * rabbits * foxes)

    delta_foxes = (-config.fox_death * foxes +
                   config.fox_growth * config.rabbit_death * rabbits * foxes)

    return [delta_rabbits, delta_foxes]


def simulate(initial_state, config, intervention=None, seed=None):
    """Simulate a run of the Lotka volterra model.

    The simulation starts at initial_state at time start_time, using the
    problem parameters specified in config. The simulator is deterministic,
    so the seed parameter is ignored.
    """
    # pylint: disable-msg=unused-argument
    def forward_map(time, state):
        """Wrap Lotka-volterra equations to use problem config."""
        # Interventions are fixed changes in dynamics after a given time.
        if intervention and time >= intervention.year:
            return dynamics(state, config.update(intervention))
        return dynamics(state, config)

    # Numerically integrate the ODE
    solution = solve_ivp(forward_map, [config.start_time, config.end_time],
                         dataclasses.astuple(initial_state))

    times = solution.t
    states = []
    for pop0, pop1 in zip(solution.y[0], solution.y[1]):
        states.append(State(pop0, pop1))
    return wn.framework.Run(states=states, times=times)



"""Experiments for world3 simulator."""
import numpy as np

from whynot.framework import DynamicsExperiment, parameter
from whynot.simulators import world3


def get_experiments():
    """Return all experiments for world3."""
    return [PollutionRCT,
            PollutionConfounding,
            PollutionUnobservedConfounding,
            PollutionMediation]


def sample_initial_states(rng):
    """Sample initial state by randomly perturbing the default state."""
    def random_scale(scale, base=10):
        """Make 10**(-scale) to 10**scale smaller/bigger uniformly."""
        return rng.choice(np.logspace(-scale, scale, base=base, num=50))

    state = world3.State()
    state.population_0_to_14 *= random_scale(0.33)
    state.population_15_to_44 *= random_scale(0.33)
    state.population_45_to_64 *= random_scale(0.33)
    state.population_65_and_over *= random_scale(0.33)
    state.industrial_capital *= random_scale(0.33)
    state.service_capital *= random_scale(0.33)
    state.arable_land *= random_scale(0.33)
    state.potentially_arable_land *= random_scale(0.33)
    state.urban_industrial_land *= random_scale(0.33)
    state.nonrenewable_resources *= random_scale(0.33)
    state.persistent_pollution *= random_scale(0.33)
    state.land_fertility *= random_scale(0.26)

    return state


##################
# RCT Experiment
##################
# pylint: disable-msg=invalid-name
#: An RCT experiment to study the effect of decreases in pollution generation on population.
PollutionRCT = DynamicsExperiment(
    name="world3_rct",
    description="Study effect of intervening in 1975 to decrease pollution generation.",
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.),
    intervention=world3.Intervention(year=1975, persistent_pollution_generation_factor=0.85),
    state_sampler=sample_initial_states,
    propensity_scorer=0.5,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=lambda run: run.initial_state.values())


##########################
# Confounding Experiments
##########################
@parameter(name="treatment_bias", default=0.8, values=np.linspace(0.5, 0.05, 1),
           description="Bias of probability of treatment between top and bottom pollution runs.")
def pollution_confounded_propensity(intervention, control_runs, treatment_bias):
    """Probability of treating each unit.

    To generate confounding, we are more likely to treat worlds with high pollution.
    """
    def persistent_pollution(run):
        return run[intervention.year].persistent_pollution

    pollution = [persistent_pollution(run) for run in control_runs]
    upper_quantile = np.quantile(pollution, 0.9)

    def treatment_prob(idx):
        if pollution[idx] > upper_quantile:
            return treatment_bias
        return 1. - treatment_bias

    return np.array([treatment_prob(idx) for idx in range(len(control_runs))])


# pylint: disable-msg=invalid-name
#: An observational experiment with confounding. Polluted states are more likely to be treated.
PollutionConfounding = DynamicsExperiment(
    name="world3_pollution_confounding",
    description=("Study effect of intervening to decrease pollution.  Confounding "
                 "arises becauses states with high pollution are more likely "
                 "to receive treatment."),
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.),
    intervention=world3.Intervention(year=1975, persistent_pollution_generation_factor=0.85),
    state_sampler=sample_initial_states,
    propensity_scorer=pollution_confounded_propensity,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=lambda intervention, run: run[intervention.year].values())


# pylint: disable-msg=invalid-name
#: An observational experiment with unobserved confounding.
PollutionUnobservedConfounding = DynamicsExperiment(
    name="world3_pollution_unobserved_confounding",
    description=("Study effect of intervening to decrease pollution.  Confounding "
                 "arises becauses states with high pollution are more likely "
                 "to receive treatment. However, only variables that directly affect "
                 "treatment assignment are observed."),
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.),
    intervention=world3.Intervention(year=1975, persistent_pollution_generation_factor=0.85),
    state_sampler=sample_initial_states,
    propensity_scorer=pollution_confounded_propensity,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=lambda intervention, run: np.array([
        run[intervention.year].persistent_pollution]))


########################
# Mediation Experiments
########################
@parameter(name="mediation_year", default=2015, values=[1980, 2000, 2020, 2045],
           description="What year to get mediating variables")
@parameter(name="num_mediators", default=11, values=[0, 3, 5, 7, 10],
           description="number of mediators to include.")
def mediation_covariates(intervention, run, mediation_year, num_mediators):
    """Build the causal dataset."""
    # Just use all of the states at the moment of treatment assignment
    confounders = run[intervention.year].values()
    mediators = run[mediation_year].values()[:num_mediators]
    return np.concatenate([confounders, mediators])


# pylint: disable-msg=invalid-name
#: An observational experiment with mediation from states after intervention.
PollutionMediation = DynamicsExperiment(
    name="world3_pollution_mediation",
    description=("Study effect of intervening to decrease pollution.  Confounding "
                 "arises becauses states with high pollution are more likely "
                 "to receive treatment. Mediation arises since future states "
                 "are also observed."),
    simulator=world3,
    simulator_config=world3.Config(persistent_pollution_generation_factor=1.),
    intervention=world3.Intervention(year=1975, persistent_pollution_generation_factor=0.85),
    state_sampler=sample_initial_states,
    propensity_scorer=pollution_confounded_propensity,
    outcome_extractor=lambda run: run[2050].total_population,
    covariate_builder=mediation_covariates)



#!/usr/bin/env python

from __future__ import print_function
from common import *
import sys
import os
import glob
import shutil
import bz2

def buildPackageUsingConda(data, fileName, version):

    tmpDir = tempfile.mkdtemp(prefix = "simpactbuildtmp")
    os.chdir(tmpDir)
    print("Work directory is " + tmpDir)

    fullFileName = os.path.join(tmpDir, fileName)
    open(fullFileName, "wb").write(data)

    os.mkdir("condapack")
    open("condapack/meta.yaml", "wt").write("""
package:
  name: simpactcyan
  version: "{}"

source:
  fn: {}
  url: file://{}

build:
  number: 0

requirements:
  build:
    - gcc 
    - cmake     
    - gsl
    - libtiff 
    - python

  run:
    - gsl
    - libtiff
        
""".format(version, fileName, fullFileName))

    open("condapack/build.sh","wt").write("""
#!/bin/bash

unset MACOSX_DEPLOYMENT_TARGET
mkdir build && cd build && CFLAGS="-I $PREFIX/include" && CXXFLAGS="-I $PREFIX/include" cmake .. -DCMAKE_INSTALL_PREFIX="$PREFIX" && make && make install
""")

    open("condapack/bld.bat", "wt").write("")

    # This builds the executables
    subprocess.check_call( ["conda", "build", "condapack" ])

    return tmpDir

def createDmg(tmpDir, version, startDir):

    # Get the conda dir
    out = S(subprocess.check_output( [ "conda", "info"]))
    condaDir = None
    for l in out.splitlines():
        if "default environment" in l:
            condaDir = l.split(" : ")[1]
            break

    if condaDir is None:
        raise Exception("Couldn't detect conda directory")

    packageFile = glob.glob(os.path.join(condaDir, "conda-bld", "osx-64", "simpactcyan-" + version + "-*.tar.bz2"))[0]
    print("Package file is " + packageFile)

    shutil.copytree(os.path.join(startDir, "platform-osx"), os.path.join(tmpDir, "packages-osx"))

    projectFileData = open(os.path.join(tmpDir, "packages-osx", "SimpactCyan.pkgproj.templ"), "rt").read()
    projectFileData = projectFileData.replace("SIMPACTCYANPROJECTBASEPATH", tmpDir)
    open(os.path.join(tmpDir, "packages-osx", "SimpactCyan.pkgproj"), "wt").write(projectFileData)

    bf = bz2.BZ2File(packageFile, mode="r")
    uncompressedData = bf.read()
    tf = tarfile.open(fileobj = StringIO(uncompressedData))

    os.chdir(tmpDir)
    tf.extractall()

    for exe in glob.glob("bin/*"):
        shutil.copy(exe, "packages-osx/Applications/SimpactCyan.app/Contents/bin/" + os.path.basename(exe))

    for dat in glob.glob("share/simpact-cyan/*"):
        if not os.path.isdir(dat):
            shutil.copy(dat, "packages-osx/Applications/SimpactCyan.app/Contents/data/" + os.path.basename(dat))

    shutil.copy("share/simpact-cyan/python/pysimpactcyan.py", os.path.join("packages-osx", "Applications", "SimpactCyan.app", "Contents", "python", "pysimpactcyan.py"))

    print("Copying executables from conda build and changing library locations")
    # Copy the used libraries from conda to the 'lib' dir
    os.chdir("packages-osx/Applications/SimpactCyan.app/Contents/bin")
    for exeName in glob.glob("*"):
        out = S(subprocess.check_output(["otool", "-L", exeName]))
        for l in out.splitlines():
            if "@rpath" in l:
                rpathLibName = l.split()[0]
                libName = os.path.basename(rpathLibName)

                shutil.copy(os.path.join(condaDir, "lib", libName), os.path.join("..", "lib", libName))
                subprocess.check_call(["install_name_tool", "-change", rpathLibName, "@executable_path/../lib/" + libName, exeName])

    print("Copying additional needed libraries from conda build")
    # Do the same for the libraries
    os.chdir("../lib")
    count = 20
    found = True
    while count > 0 and found:
        found = False
        count -= 1
        for exeName in glob.glob("*"):
            out = S(subprocess.check_output(["otool", "-L", exeName]))
            for l in out.splitlines():
                if "@rpath" in l and not exeName in l:
                    rpathLibName = l.split()[0]
                    libName = os.path.basename(rpathLibName)

                    if not os.path.exists(libName):
                        shutil.copy(os.path.join(condaDir, "lib", libName), libName)
                    
                    found = True
                    subprocess.check_call(["install_name_tool", "-change", rpathLibName, "@executable_path/../lib/" + libName, exeName])

    os.chdir(os.path.join(tmpDir, "packages-osx"))
    subprocess.check_call("./createiconfile.sh")
    subprocess.check_call("./addicon.sh")
    subprocess.check_call(["packagesbuild", "SimpactCyan.pkgproj"])
    subprocess.check_call(["./addicon2.sh"])

    os.chdir("build")
    subprocess.check_call(["hdiutil", "create", "SimpactCyan-{}.dmg".format(version), "-volname", "SimpactCyan-{}".format(version), "-fs", "HFS+", "-srcfolder", "SimpactCyan.mpkg"])

    destFile = os.path.join(os.getcwd(), "SimpactCyan-{}.dmg".format(version))
    return destFile

def main():

    startDir = os.getcwd()
    if not os.path.exists("platform-osx"):
        raise Exception("Need to be started from the directory which has 'platform-osx' as subdirectory")

    try:
        fullFileName = sys.argv[1]
        if len(sys.argv) > 2:
            raise Exception("Too many arguments")
    except Exception as e:
        print("""Error: {}

Usage: {} path

Here 'path' can be a local .tar.gz file or a .tar.gz file available over http.
""".format(e, sys.argv[0]))
        sys.exit(-1)

    try:
        print("Checking if 'packagesbuild' exists")
        subprocess.check_call(["packagesbuild"])
    except subprocess.CalledProcessError:
        pass
    except Exception as e:
        print("""Error: {}

Couldn't run 'packagesbuild', which is needed to create a dmg file.
Install http://s.sudre.free.fr/Software/Packages/about.html
""".format(e))
        sys.exit(-1)

    try:
        print("Checking if 'conda' can be run")
        subprocess.check_output([ "conda", "--version" ])
    except OSError as e:
        print("""Error: {}

The OS X package creation program relies on the Anaconda Python environment
to build the executables (to be able to use gcc instead of clang). Make sure 
this is installed and that the 'conda' command can be run from the command 
line.
""".format(e))
        sys.exit(-1)

    subprocess.call(["conda", "config", "--add", "channels", "http://conda.anaconda.org/jori"])
    subprocess.call(["conda", "install", "-y", "conda-build"])
    subprocess.call(["conda", "install", "-y", "gsl"])
    subprocess.call(["conda", "install", "-y", "libtiff"])
    subprocess.call(["conda", "install", "-y", "gcc"])

    fileName = os.path.basename(fullFileName)
    version = getVersionFromFilename(fileName)

    print("Version: " + version)

    data = getPackageData(fullFileName)
    tmpDir = buildPackageUsingConda(data, fileName, version)

    destFile = createDmg(tmpDir, version, startDir)
    print("")
    print("Package is " + destFile)


if __name__ == "__main__":
    main()



"""World 2 dynamics simulator."""
import copy
import dataclasses

import numpy as np

import whynot as wn
from whynot.simulators.world2 import tables


@dataclasses.dataclass
class Config:
    # pylint: disable-msg=too-few-public-methods
    """Dynamics and simulation parameters for World2.

    Default values correspond to the standard run of World2.

    Examples
    --------
    >>> # Set the initial birth rate to 0.05
    >>> world2.config(birth_rate=0.05)

    """

    # The #: comments allow Sphinx to autodocument these parameters.
    #:
    birth_rate: float = 0.04
    #:
    death_rate: float = 0.028
    #:
    effective_capital_investment_ratio: float = 1.0
    #:
    natural_resources_usage: float = 1.0
    #:
    land_area: float = 135e6
    #:
    population_density: float = 26.5
    #:
    food_coefficient: float = 1.0
    #:
    food_normal: float = 1.0
    #:
    capital_investment_agriculture: float = 0.3
    #:
    capital_investment_generation: float = 0.05
    #:
    capital_investment_discard: float = 0.025
    #:
    pollution_standard: float = 3.6e9
    #:
    pollution: float = 1.0
    #:
    intervention_time: float = 1970
    #:
    capital_investment_in_agriculture_frac_adj_time: float = 15.0
    #:
    quality_of_life_standard: float = 1.0

    #: Year to initialize the simulation.
    start_time: float = 1900
    #: Year to end the simulation.
    end_time: float = 2100
    #: Time elapsed on each update of the discrete dynamics (forward euler).
    delta_t: float = 0.2

    def update(self, intervention):
        """Create a new config after applying the intervention."""
        intervention = dataclasses.asdict(intervention)
        del intervention["year"]
        updates = dict((param, value) for param, value in intervention.items() if value is not None)
        return dataclasses.replace(self, **updates)


@dataclasses.dataclass
class Intervention:
    # pylint: disable-msg=too-few-public-methods
    """Class to specify Interventions in World2 dynamics.

    By default, all parameters are initialized to `None`. Only parameters
    that are explicitly initialized to a non-`None` value will be intervened upon.

    Examples
    --------
    >>> # Set the death rate to 0.06 in 1980. All other parameters unchanged.
    >>> world2.Intervention(year=1980, death_rate=0.06)

    .. note::
       For consistency with the original world2, only a subset of the dynamics
       parameters can be intervened upon.

    """

    #: Year to perform the intervention.
    year: float = 1970

    # The default value of all variables is None, which means no change
    # in the corresponding variable after intervention.

    #: birth_rate after intervention.
    birth_rate: float = None
    #: death_rate after intervention.
    death_rate: float = None
    #: natural_resource_usage rate after intervention.
    natural_resources_usage: float = None
    #: food_coefficient after intervention.
    food_coefficient: float = None
    #: investment_discard rate after intervention.
    capital_investment_discard: float = None
    #: investment_generation rate after intervention.
    capital_investment_generation: float = None
    #: pollution rate after intervention.
    pollution: float = None


@dataclasses.dataclass
class State:
    # pylint: disable-msg=too-few-public-methods
    """World 2 state.

    Default values correspond to the initial state in the standard run of World2.

    """

    #: Total population stock
    population: float = 1.65e9
    #: Stock of natural resources
    natural_resources: float = 3.6e9 * 250
    #: Total capital investment stock
    capital_investment: float = 0.4e9
    #: Pollution stock
    pollution: float = 0.2e9
    #: Fraction of capital investment in agriculture
    capital_investment_in_agriculture: float = 0.2
    #: (Derived) quality of life metric
    quality_of_life: float = 1.0

    def values(self):
        """Return the state as a numpy array."""
        state_dict = dataclasses.asdict(self)
        return np.array([state_dict[key] for key in sorted(state_dict.keys())])


#################
# Dynamics
#################
class WorldDynamics():
    """Class encapsulating Forrester world2 dynamics."""

    def __init__(self, config, intervention, initial_state):
        """Initialize world2 dynamics at initial_state with config."""
        self._state = copy.deepcopy(initial_state)
        self.initial_natural_resources = initial_state.natural_resources
        self.intervention = intervention

        config = copy.deepcopy(config)
        self.preintervention_config = config
        if intervention:
            self.postintervention_config = config.update(intervention)
        else:
            self.postintervention_config = config
        self.config = config

    def __str__(self):
        """Print all of the state variables."""
        template = "Population: {}\nNatural Resources: {}\nCapital Investment: {}\n"
        template += "Pollution: {}\nFraction Capital Investment in Agriculture: {}\n"

        return template.format(*[f.name for f in dataclasses.fields(self._state)])

    @property
    def state(self):
        """Return a copy of the current world2 state."""
        return copy.deepcopy(self._state)

    @property
    def capital_investment_ratio(self):
        """Return current capital investment ratio."""
        return self._state.capital_investment / self._state.population

    @property
    def crowding_ratio(self):
        """Return current crowding ratio."""
        return self._state.population / (self.config.land_area * self.config.population_density)

    @property
    def pollution_ratio(self):
        """Return current pollution ratio."""
        return self._state.pollution / self.config.pollution_standard

    @property
    def food_ratio(self):
        """Return current food ratio."""
        capital_investment_ratio_in_agr = (self.capital_investment_ratio
                                           * self._state.capital_investment_in_agriculture
                                           / self.config.capital_investment_agriculture)

        pollution_multiplier = tables.FOOD_FROM_POLLUTION[self.pollution_ratio]
        crowding_multiplier = tables.FOOD_FROM_CROWDING[self.crowding_ratio]
        capital_investment_food_potential = \
            tables.FOOD_POTENTIAL_FROM_CAPITAL_INVESTMENT[capital_investment_ratio_in_agr]

        food_ratio = (capital_investment_food_potential
                      * crowding_multiplier
                      * pollution_multiplier
                      * self.config.food_coefficient
                      / self.config.food_normal)

        return food_ratio

    @property
    def standard_of_living(self):
        """Return the current standard of living."""
        # Natural resources
        frac_natural_resources_remaining = (self._state.natural_resources /
                                            self.initial_natural_resources)
        natural_resources_extraction_multiplier = \
            tables.NATURAL_RESOURCE_EXTRACTION[frac_natural_resources_remaining]

        effective_capital_investment_ratio = (
            self.capital_investment_ratio
            * natural_resources_extraction_multiplier
            * (1.0 - self._state.capital_investment_in_agriculture)
            / (1.0 - self.config.capital_investment_agriculture))

        return effective_capital_investment_ratio / self.config.effective_capital_investment_ratio

    @property
    def death_rate_per_year(self):
        """Return the fraction of people who die in the current year."""
        material_multiplier = tables.DEATH_RATE_FROM_MATERIAL[self.standard_of_living]
        pollution_multiplier = tables.DEATH_RATE_FROM_POLLUTION[self.pollution_ratio]
        food_multiplier = tables.DEATH_RATE_FROM_FOOD[self.food_ratio]
        crowding_multiplier = tables.DEATH_RATE_FROM_CROWDING[self.crowding_ratio]

        death_rate_per_year = (
            self.config.death_rate
            * material_multiplier
            * pollution_multiplier
            * food_multiplier
            * crowding_multiplier)

        return death_rate_per_year

    @property
    def birth_rate_per_year(self):
        """Return the fraction of people born in the current year."""
        material_multiplier = tables.BIRTH_RATE_FROM_MATERIAL[self.standard_of_living]
        pollution_multiplier = tables.BIRTH_RATE_FROM_POLLUTION[self.pollution_ratio]
        food_multiplier = tables.BIRTH_RATE_FROM_FOOD[self.food_ratio]
        crowding_multiplier = tables.BIRTH_RATE_FROM_CROWDING[self.crowding_ratio]

        birth_rate_per_year = (
            self.config.birth_rate
            * material_multiplier
            * pollution_multiplier
            * food_multiplier
            * crowding_multiplier)

        return birth_rate_per_year

    @property
    def natural_resources_usage_rate(self):
        """Return the rate of natural resource usage at the current time."""
        material_multiplier = tables.NATURAL_RESOURCES_FROM_MATERIAL[self.standard_of_living]

        natural_resources_usage_rate = (
            self._state.population
            * self.config.natural_resources_usage
            * material_multiplier)

        return natural_resources_usage_rate

    @property
    def capital_investment_rate(self):
        """Return the rate of capital investment at the current time."""
        material_multiplier = tables.CAPITAL_INVESTMENT_MULTIPLIER_TABLE[self.standard_of_living]

        capital_investment_generation = (
            self._state.population
            * material_multiplier
            * self.config.capital_investment_generation)

        capital_investment_discard = (
            self._state.capital_investment
            * self.config.capital_investment_discard)

        return capital_investment_generation - capital_investment_discard

    @property
    def pollution_rate(self):
        """Return the current rate of pollution."""
        capital_multiplier = tables.POLLUTION_FROM_CAPITAL[self.capital_investment_ratio]

        pollution_generation = (
            self._state.population
            * capital_multiplier
            * self.config.pollution)

        absorption_time = tables.POLLUTION_ABSORPTION_TIME_TABLE[self.pollution_ratio]
        pollution_absorption = self._state.pollution / absorption_time

        return pollution_generation - pollution_absorption

    @property
    def quality_of_life(self):
        """Return the current quality of life meta-statistic."""
        material_multiplier = tables.QUALITY_OF_LIFE_FROM_MATERIAL[self.standard_of_living]
        crowding_multiplier = tables.QUALITY_OF_LIFE_FROM_CROWDING[self.crowding_ratio]
        food_multiplier = tables.QUALITY_OF_LIFE_FROM_FOOD[self.food_ratio]
        pollution_multiplier = tables.QUALITY_OF_LIFE_FROM_POLLUTION[self.pollution_ratio]

        quality_of_life = (
            self.config.quality_of_life_standard
            * material_multiplier
            * crowding_multiplier
            * food_multiplier
            * pollution_multiplier)

        return quality_of_life

    def capital_invest_agr_frac_delta(self, delta_t):
        """Return the change in the fraction of captial invested in agriculture."""
        # I have no intuition for this formula
        capital_fraction_indicated_by_food_ratio =\
            tables.CAPITAL_FRACTION_INDICATE_BY_FOOD_RATIO_TABLE[self.food_ratio]

        quality_of_life_material = tables.QUALITY_OF_LIFE_FROM_MATERIAL[self.standard_of_living]
        quality_of_life_food = tables.QUALITY_OF_LIFE_FROM_FOOD[self.food_ratio]

        life_quality_ratio = quality_of_life_material / quality_of_life_food
        capital_investment_from_quality_ratio = \
            tables.CAPITAL_INVESTMENT_FROM_QUALITY[life_quality_ratio]

        delta = (delta_t / self.config.capital_investment_in_agriculture_frac_adj_time) \
            * (capital_fraction_indicated_by_food_ratio
               * capital_investment_from_quality_ratio
               - self._state.capital_investment_in_agriculture)

        return delta

    def step(self, time, delta_t):
        """Advance the simulation forward delta_t steps, starting at time TIME."""
        if self.intervention and time >= self.intervention.year:
            self.config = self.postintervention_config
        else:
            self.config = self.preintervention_config

        # Population
        deaths = self.death_rate_per_year * self._state.population * delta_t
        births = self.birth_rate_per_year * self._state.population * delta_t
        population_delta = births - deaths

        # Natural resources (negative since this is usage)
        natural_resources_delta = -self.natural_resources_usage_rate * delta_t

        # Capital_investment
        capital_investment_delta = self.capital_investment_rate * delta_t

        # Pollution
        pollution_delta = self.pollution_rate * delta_t

        # Investment in agriculture
        capital_investment_in_agr_frac_delta = self.capital_invest_agr_frac_delta(delta_t)

        # Update the state variables. Gather deltas first to ensure the updates
        # are atomic, e.g. since changing the population will also change the
        # resources delta.
        self._state.population += population_delta
        self._state.natural_resources += natural_resources_delta
        self._state.capital_investment += capital_investment_delta
        self._state.pollution += pollution_delta
        self._state.capital_investment_in_agriculture += capital_investment_in_agr_frac_delta
        self._state.quality_of_life = self.quality_of_life

    def apply_shock(self, shock):
        """Apply an exogenous shock to the current state of the simulator."""
        self._state.population += shock.population
        self._state.natural_resources += shock.natural_resources
        self._state.capital_investment += shock.capital_investment
        self._state.pollution += shock.pollution
        self._state.capital_investment_in_agriculture += shock.capital_investment_in_agriculture


def simulate(initial_state, config, intervention=None, exogenous_shocks=None, seed=None):
    # pylint: disable=unused-argument
    """Run a simulation of the world2 dynamics, starting at initial_state at time end_time.

    The system of ODEs is solved using the forward-euler method with
    discretization delta_t.

    The user can optionally supply a list of exogenous shocks to the state
    variables.  Exogenous shocks should be a tuple (time, shock),
    where shock is a State object denoting the change in state to implement
    at the specific time.

    The simulator is current deterministic, and the SEED parameter is ignored.

    Returns a list of states objects, and a list of times, where
    states[i] is the state of the simulator at when measured at time times[i].
    """
    if exogenous_shocks is None:
        exogenous_shocks = []

    world = WorldDynamics(config, intervention, initial_state)

    states = []
    times = np.arange(config.start_time, config.end_time + config.delta_t, config.delta_t)
    for time in times:
        states.append(world.state)
        world.step(time, config.delta_t)
        # Hacky addition to get exogenous shocks
        for shock_time, shock in exogenous_shocks:
            if np.isclose(time, shock_time):
                world.apply_shock(shock)
    states.append(world.state)
    return wn.framework.Run(states=states, times=times)



"""Test suite for whynot framework."""
import copy
import dataclasses

import numpy as np
import pytest
import statsmodels

import whynot as wn
from whynot.framework import parameter

def test_framework_run():
    """Test the whynot.Run object"""
    states = ["start_state", 2, 3, 4]
    times = [4, 5, 6, 10]

    run = wn.framework.Run(states=states, times=times)

    assert run.initial_state == "start_state"
    assert run[6] == 3


def test_parameter():
    """Test the parameter decorator."""
    #Ccase when no standard arguments
    @parameter(name="a", default=1, values=[1, 2], description="test a")
    @parameter(name="b", default=3, values=[3, 4], description="test b")
    def foo(a, b):
        return a + b

    def check_params(listed, returned):
        assert set(listed) == set(p.name for p in returned)

    check_params(["a", "b"], wn.framework.extract_params(foo, standard_args=[]))

    @parameter(name="b", default=3, values=[3, 4], description="test b")
    def bar(a, b):
        return a + b

    # Can omit a as a parameter if a is one of the "standard args"
    check_params(["b"], wn.framework.extract_params(bar, standard_args=["a"]))

    # If there are no standard args, raise a value error since "a" is not
    # listed as a parameter
    with pytest.raises(ValueError):
        wn.framework.extract_params(bar, standard_args=[])

    # If a parameter is added, but omitted form the signature, we should raise
    with pytest.raises(ValueError):
        @parameter(name="a", default=0)
        def baz(propensity):
            return propensity

    # If a parameter is also part of standard args, we should raise
    with pytest.raises(ValueError):
        @parameter(name="rng", default=0)
        def foobar(rng):
            return 2
        wn.framework.extract_params(foobar, standard_args=["rng"])


def test_parameter_collection():
    """Test the parameter collection object."""
    from whynot.framework import ExperimentParameter

    p1 = ExperimentParameter(name="p1", default=1)
    collection = wn.framework.ParameterCollection(params=[p1])
    collection.add_parameter(ExperimentParameter(name="p2", default=2))

    projected = collection.project({"p1": 10})
    assert projected["p1"] == 10
    assert projected["p2"] == 2


    p3 = ExperimentParameter(name="p3", default=3)
    p4 = ExperimentParameter(name="p4", default=4)
    collection2 = wn.framework.ParameterCollection(params=[p3, p4])

    collection += collection2

    assert collection["p3"].default == 3

    # Ensure names are unique
    a = ExperimentParameter(name="a", default=3)
    aa = ExperimentParameter(name="a", default=4)
    with pytest.raises(ValueError):
        wn.framework.ParameterCollection(params=[a, aa])

    # Ensure names are unique after adding parameter
    with pytest.raises(ValueError):
        coll = wn.framework.ParameterCollection(params=[a])
        coll.add_parameter(aa)

    # Ensure names are unique if you add two collections
    with pytest.raises(ValueError):
        coll_one = wn.framework.ParameterCollection(params=[a])
        coll_two = wn.framework.ParameterCollection(params=[aa])
        new = coll_one + coll_two


#################################
# Dynamics experiment class tests
#################################
@dataclasses.dataclass
class ToyConfig:
    num_steps: int = 10

@dataclasses.dataclass
class Intervention(ToyConfig):
    year: int = 0
    effect_size: int = 1

@dataclasses.dataclass
class ToyState:
    val: int = 0

class ToySimulator():
    """Simple simulator for testing purposes.

    Dynamics are to add effect_size to state at each time step if
    the run is treated. Otherwise, the run is constant starting at
    the initial_state value.

    """

    @staticmethod
    def simulate(initial_state, config, intervention=None, seed=None):
        """If treated, add a constant offset at each timestep."""
        states = []
        times = list(range(config.num_steps))

        state = initial_state
        for time in times:
            states.append(copy.deepcopy(state))
            if intervention and time >= intervention.year:
                state.val += intervention.effect_size
        return wn.framework.Run(states=states, times=times)

TEST_SIMULATOR = ToySimulator()

def check_shapes(dataset, num_samples, num_features):
    """Ensure the output has the right size."""
    assert dataset.covariates.shape == (num_samples, num_features)
    assert dataset.treatments.shape == (num_samples,)
    assert dataset.outcomes.shape == (num_samples,)
    assert dataset.true_effects.shape == (num_samples,)


def check_rct_assigment(assignment, prob):
    """Ensure that RCT assignment has approximately the correct proportion of treated."""
    total_treated = np.sum(assignment)
    ci_low, ci_upp = statsmodels.stats.proportion.proportion_confint(
        count=total_treated, nobs=len(assignment), alpha=0.001, method="beta")
    assert ci_low <= prob <= ci_upp

### Helper methods for DynamicsExperiment tests
def basic_state_sampler():
    """Initial state is 0 for all rollouts."""
    return ToyState(val=0)

def correlated_state_sampler(num_samples):
    """Sample correlated initial states."""
    states = []
    for sample_idx in range(num_samples):
        states.append(ToyState(val=sample_idx))
    return states

def basic_outcome_extractor(run):
    """Look at state value after 5 steps."""
    return run[5].val

def basic_covariate_builder(run):
    """Return the final state of the run."""
    return np.array(run.states[-1].val)

def build_experiment(state_sampler=basic_state_sampler,
                     config=ToyConfig(num_steps=10),
                     intervention=Intervention(year=0, effect_size=1),
                     propensity_scorer=0.5,
                     outcome_extractor=basic_outcome_extractor,
                     covariate_builder=basic_covariate_builder):
    """Construct a dynamics experiment instance with sane defaults."""
    return wn.framework.DynamicsExperiment(
        name="basic_test",
        description="Sanity check.",
        simulator=TEST_SIMULATOR,
        simulator_config=config,
        intervention=intervention,
        state_sampler=state_sampler,
        propensity_scorer=propensity_scorer,
        outcome_extractor=outcome_extractor,
        covariate_builder=covariate_builder)

# DynamicsExperiment class tests.
def test_dynamics_experiment_basic():
    """Sanity check: does the class work?"""

    basic_exp = build_experiment()

    # Check outputs are the right sizes
    dataset = basic_exp.run(num_samples=10)
    check_shapes(dataset, 10, 1)

    dataset = basic_exp.run(num_samples=200)
    check_shapes(dataset, 200, 1)


    # Make sure the true effect is correctly computed
    assert np.allclose(5, dataset.true_effects)

    # Make sure the assignment and outcome are consistent
    assert np.allclose(0, dataset.outcomes[dataset.treatments == 0.])
    assert np.allclose(5, dataset.outcomes[dataset.treatments == 1.])

    # Make sure that incorrect parameters are flagged
    with pytest.raises(ValueError):
        basic_exp.run(num_samples=20, parameter=1234)


def test_state_sampler():
    """Test the state sampler function."""

    def initial_state_covariates(run):
        """Outcome is the initial state of the run."""
        return run.initial_state.val

    # Ensure the state sampler correctly uses randomness
    def random_initial_state(rng):
        """Initial state is 0 for all rollouts."""
        if rng.uniform() < 0.5:
            return ToyState(val=0)
        return ToyState(val=1)

    exp = build_experiment(state_sampler=random_initial_state,
                           covariate_builder=initial_state_covariates)
    dataset0 = exp.run(num_samples=100, seed=1234)
    dataset1 = exp.run(num_samples=100, seed=1234)

    assert np.allclose(dataset0.covariates, dataset1.covariates)

   # Ensure the state sampler correctly deals with correlated state sampling
    exp = build_experiment(state_sampler=correlated_state_sampler,
                           covariate_builder=initial_state_covariates)
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(np.expand_dims(np.arange(100), axis=1), dataset.covariates)

    # Test that parameters play nicely with state sampler
    @parameter(name='init_val', default=2)
    def parameterized_sampler(init_val):
        return ToyState(val=init_val)

    exp = build_experiment(state_sampler=parameterized_sampler,
                           covariate_builder=initial_state_covariates)

    # Ensure default works
    dataset = exp.run(num_samples=10)
    assert np.allclose(dataset.covariates, 2)

    # Ensure varying parameterization works
    dataset = exp.run(num_samples=10, init_val=3.14)
    assert np.allclose(dataset.covariates, 3.14)

    dataset = exp.run(num_samples=10, init_val=1234)
    assert np.allclose(dataset.covariates, 1234)

    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_val=123)

def test_config():
    """Test simulator config methods."""
    # Make sure config is actually effective
    def covariate_builder(run):
        return len(run.times)

    config = ToyConfig(num_steps=20)
    exp = build_experiment(config=config, covariate_builder=covariate_builder)
    dataset = exp.run(num_samples=15)
    assert np.allclose(20, dataset.covariates)

    # Check that config can be a function
    config = lambda: ToyConfig(num_steps=23)
    exp = build_experiment(config=config, covariate_builder=covariate_builder)
    dataset = exp.run(num_samples=15)
    assert np.allclose(23, dataset.covariates)

    # Check that these functions can be parameterized
    @parameter(name="num_steps", default=10)
    def pconfig(num_steps):
        return ToyConfig(num_steps=num_steps)

    exp = build_experiment(config=pconfig, covariate_builder=covariate_builder)

    # Default
    dataset = exp.run(num_samples=10)
    assert np.allclose(10, dataset.covariates)

    # Vary the parameter.
    dataset = exp.run(num_samples=10, num_steps=7)
    assert np.allclose(7, dataset.covariates)
    dataset = exp.run(num_samples=10, num_steps=56)
    assert np.allclose(56, dataset.covariates)

    # Check that bad argument flagging works for configs.
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_arg=123)


def test_intervention():
    """Test intervention methods."""
    intervention = Intervention(effect_size=2)
    exp = build_experiment(intervention=intervention)
    dataset = exp.run(num_samples=20)
    assert np.allclose(dataset.true_effects, 10)

    intervention = lambda: Intervention(year=1)
    exp = build_experiment(intervention=intervention)
    dataset = exp.run(num_samples=20)
    assert np.allclose(dataset.true_effects, 4)

    @parameter(name="effect_size", default=2)
    def pintervention(effect_size):
        return Intervention(effect_size=effect_size)
    exp = build_experiment(intervention=pintervention)

    # Default
    dataset = exp.run(num_samples=20)
    assert np.allclose(dataset.true_effects, 10)

    # Varying parameter
    dataset = exp.run(num_samples=20, effect_size=3)
    assert np.allclose(dataset.true_effects, 15)
    dataset = exp.run(num_samples=20, effect_size=5)
    assert np.allclose(dataset.true_effects, 25)

    # Check that bad argument flagging works for interventions
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_arg=123)


def test_propensity_scorer_rct():
    """Basic tests of propensity scorer functionality in RCT setting."""

    # Test to ensure propensity is (approximately) correct when using a constant
    exp = build_experiment(propensity_scorer=0.3)
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.3)

    exp = build_experiment(propensity_scorer=0.8)
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.8)


    # Make sure providing a function for the propensity score also works
    exp = build_experiment(propensity_scorer=lambda: 0.3)
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.3)

    # Only floats between 0 and 1 are valid constant propensities
    with pytest.raises(ValueError):
        exp = build_experiment(propensity_scorer="random_string")
        exp.run(num_samples=1000, seed=1234)

    with pytest.raises(ValueError):
        exp = build_experiment(propensity_scorer=2.)
        exp.run(num_samples=1000, seed=1234)

    # Test parameterizing the propensity scorer
    @parameter(name="propensity", default=0.2)
    def parameterized_scorer(propensity):
        return propensity

    exp = build_experiment(propensity_scorer=parameterized_scorer)

    # Check default
    dataset = exp.run(num_samples=1000, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)

    # Check using parameter
    dataset = exp.run(num_samples=1000, seed=1234, propensity=0.65)
    check_rct_assigment(dataset.treatments, 0.65)

    # Check error handing
    with pytest.raises(ValueError):
        exp.run(num_samples=10, seed=1234, random_arg=1234)

    # Tests specifying control_config and treatment_config
    config = ToyConfig(num_steps=13)
    def config_scorer(config):
        assert config.num_steps == 13
        return 0.2
    exp = build_experiment(propensity_scorer=config_scorer, config=config)
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)

    intervene = Intervention(year=4)
    def intervention_scorer(intervention):
        assert intervention.year == 4
        return 0.2
    exp = build_experiment(propensity_scorer=intervention_scorer,
                           intervention=intervene)
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)

    def both_scorer(config, intervention):
        assert config.num_steps == 13
        assert intervention.year == 4
        return 0.2
    exp = build_experiment(propensity_scorer=both_scorer, config=config,
                           intervention=intervene)
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.2)

def test_propensity_scorer_observation():
    """Test propensity scorer in the observational setting."""

    # Test that specificying control_run, treatment_run gives you the correct
    # runs from the model.
    def crun_scorer(control_run):
        # Control runs are constant
        start = control_run.initial_state.val
        assert np.allclose(np.array([s.val for s in control_run.states]), start)
        return 0.5

    exp = build_experiment(state_sampler=correlated_state_sampler, propensity_scorer=crun_scorer)
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.5)

    def run_scorer(run):
        # Run should default to control run
        start = run.initial_state.val
        assert np.allclose(np.array([s.val for s in run.states]), start)
        return 0.5

    exp = build_experiment(state_sampler=correlated_state_sampler, propensity_scorer=run_scorer)
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.5)

    def trun_scorer(treatment_run):
        start = treatment_run.initial_state.val
        run_vals = np.array([s.val for s in treatment_run.states])
        assert np.allclose(start + np.arange(10), run_vals)
        return 0.5

    exp = build_experiment(state_sampler=correlated_state_sampler, propensity_scorer=trun_scorer)
    dataset = exp.run(num_samples=100, seed=1234)
    check_rct_assigment(dataset.treatments, 0.5)

    # Check propensity scores with interference works
    def cruns_scorer(control_runs):
        # Make sure that we see all of the runs
        # control runs should start with 1, 2, 3, 4, ..., num_samples
        start_vals = sorted([run.initial_state.val for run in control_runs])
        assert np.allclose(np.arange(len(control_runs)), start_vals)

        # Make sure they are control runs (i.e. all constant)
        for run in control_runs:
            assert np.allclose(np.array([s.val for s in run.states]), run.initial_state.val)

        # Only treat first half of runs
        n = len(control_runs)
        propensities = np.zeros((n, ))
        propensities[: n // 2] = 1.
        return propensities

    n = 100
    exp = build_experiment(state_sampler=correlated_state_sampler, propensity_scorer=cruns_scorer)
    dataset = exp.run(num_samples=n, seed=1234)
    assert np.allclose(dataset.treatments[: n // 2], 1)
    assert np.allclose(dataset.treatments[n // 2 :], 0)

    def runs_scorer(runs):
        # Runs should be the same as control runs
        return cruns_scorer(runs)

    exp = build_experiment(state_sampler=correlated_state_sampler, propensity_scorer=runs_scorer)
    dataset = exp.run(num_samples=n, seed=1234)
    assert np.allclose(dataset.treatments[: n // 2], 1)
    assert np.allclose(dataset.treatments[n // 2 :], 0)

    def truns_scorer(treatment_runs):
        # Make sure that we see all of the runs
        # control runs should start with 1, 2, 3, 4, ..., num_samples
        start_vals = sorted([run.initial_state.val for run in treatment_runs])
        assert np.allclose(np.arange(len(treatment_runs)), start_vals)

        # Make sure they are control runs (i.e. all constant)
        for run in treatment_runs:
            assert np.allclose(np.array([s.val for s in run.states]),
                               run.initial_state.val + np.arange(10))

        # Only treat first half of runs
        n = len(treatment_runs)
        propensities = np.zeros((n, ))
        propensities[: n // 2] = 1.
        return propensities

    exp = build_experiment(state_sampler=correlated_state_sampler, propensity_scorer=truns_scorer)
    dataset = exp.run(num_samples=n, seed=1234)
    assert np.allclose(dataset.treatments[: n // 2], 1)
    assert np.allclose(dataset.treatments[n // 2 :], 0)


def test_outcome_extractor():
    """Test the outcome extractor."""

    def compute_outcome(run):
        """Look at state value after 5 steps."""
        return run[5].val

    # Make sure compute outcome works as expected
    exp = build_experiment(outcome_extractor=compute_outcome)
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(dataset.outcomes[dataset.treatments == 0], 0)
    assert np.allclose(dataset.outcomes[dataset.treatments == 1], 5)

    # Compose compute_outcome with parameters
    @parameter(name='observation_step', default=3)
    def parameter_compute_outcome(run, observation_step):
        """Look at state value after observation_step steps."""
        return run[observation_step].val

    exp = build_experiment(outcome_extractor=parameter_compute_outcome)

    # Check defaults
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(dataset.outcomes[dataset.treatments == 0], 0)
    assert np.allclose(dataset.outcomes[dataset.treatments == 1], 3)

    # Check parameterized version
    dataset = exp.run(num_samples=100, seed=1234, observation_step=8)
    assert np.allclose(dataset.outcomes[dataset.treatments == 0], 0)
    assert np.allclose(dataset.outcomes[dataset.treatments == 1], 8)

    # Make sure error checking works
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_param=123)


def test_covariate_builder():
    """Simple tests for covariate builder function."""

    @parameter(name="step", default=4)
    def build_covariates(config, intervention, run, step):
        """Return the final state of the run."""
        # Make sure we're getting the correct config/intervention
        assert config.num_steps == 10 and intervention.effect_size == 1
        return run.states[step].val

    exp = build_experiment(covariate_builder=build_covariates)

    # Check defaults
    dataset = exp.run(num_samples=100, seed=1234)
    assert np.allclose(dataset.covariates[dataset.treatments == 0], 0)
    assert np.allclose(dataset.covariates[dataset.treatments == 1], 4)

    # Check parameter
    dataset = exp.run(num_samples=100, seed=1234, step=7)
    assert np.allclose(dataset.covariates[dataset.treatments == 0], 0)
    assert np.allclose(dataset.covariates[dataset.treatments == 1], 7)

    # Make sure error checking works
    with pytest.raises(ValueError):
        exp.run(num_samples=10, random_param=123)



"""Unit tests for Dice simulator."""
import os

import numpy as np
import pandas as pd

import whynot as wn


def test_gams():
    """Ensure calling default simulator has results reflective of the GAMS implementaiton"""
    # Load results from standard GAMS results
    gams_file = os.path.realpath(__file__).rsplit('/', 1)[0] + '/gamsBaseline.csv'
    gams = pd.read_csv(gams_file).replace([np.inf, -np.inf], np.nan).fillna(0)
    gams.columns = ['Variable', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10']
    gams.drop("Variable", axis=1, inplace=True)

    # Compute simulator results
    config = wn.dice.Config(numPeriods=10, ifopt=0)
    initial_state = wn.dice.State()
    run = wn.dice.simulate(initial_state, config, stochastic=False)

    # Run without randomness over the config parameters.
    sim_dict = {var: [getattr(state, var) for state in run.states]
                for var in initial_state.variables}
    sim = pd.DataFrame.from_dict(sim_dict, orient='index',
                                 columns=gams.columns).replace([np.inf, -np.inf], np.nan).fillna(0)

    np.testing.assert_allclose(sim, gams, rtol=1e-4, atol=5e-2, equal_nan=False)

if __name__ == "__main__":
    test_gams()



#!/usr/bin/env python

import subprocess
import json
import pprint
import glob
import sys

paramNamesConfig = set()
for exe in ["simpact-cyan-release", "maxart-release" ]:
    data = subprocess.check_output([exe, "--showconfigoptions"])
    opts = json.loads(data)

    conf = opts["configNames"]
    for n in conf:
        params = conf[n]["params"]
        for p in params:
            paramNamesConfig.add(p[0])

dist = opts["distTypes"]
for d in dist:
    paramNamesConfig.add("some.option.dist")
    for p in dist[d]["params"]:
        paramNamesConfig.add("some.option.dist." + d + "." + p[0])

dist = opts["distTypes2D"]
for d in dist:
    paramNamesConfig.add("some.option.dist2d")
    for p in dist[d]["params"]:
        paramNamesConfig.add("some.option.dist2d." + d + "." + p[0])

paramNamesDoc = set()
data = [ ]
for n in glob.glob("*.rst"):
    data += open(n).readlines()

for l in data:
    l = l.strip()
    if l.startswith("- ``"):
        if l.endswith(": |br|"):
            parts = l.split('`')
            name = parts[2]
            suff = [ (".dist.type", ".dist"), (".dist2d.type", ".dist2d") ]
            for s in suff:
                if name.endswith(s[0]):
                    name = name[:-len(s[0])] + s[1]
            
            paramNamesDoc.add(name)

diff1 = paramNamesConfig-paramNamesDoc
diff2 = paramNamesDoc-paramNamesConfig

diff1 = sorted([ n for n in diff1 ])
diff2 = sorted([ n for n in diff2 ])

print "In config file but not in documentation"
for n in diff1:
    print "  " + n
print

print "In documentation but not in config file"
for n in diff2:
    print "  " + n



"""Experiments for world2 simulator."""
import numpy as np

from whynot.framework import DynamicsExperiment, parameter
from whynot.simulators import world2


def get_experiments():
    """Return all experiments for world2."""
    return [RCT, BiasedTreatment, Mediation]


def sample_initial_states(rng):
    """Sample an initial world2 state by perturbing the default."""
    state = world2.State()
    state.population *= rng.uniform(0.75, 2)
    state.natural_resources *= rng.uniform(0.25, 10.)
    state.capital_investment *= rng.uniform(0.5, 2.0)
    state.pollution *= rng.uniform(0.5, 2)
    state.capital_investment_in_agriculture *= rng.uniform(0.5, 1.5)
    return state


###########################
# RCT Experiment
###########################
#: RCT experiment for world2.
RCT = DynamicsExperiment(
    name="world2_rct",
    description="RCT to measure the effect of increasing capital investment in 1970.",
    simulator=world2,
    simulator_config=world2.Config(),
    intervention=world2.Intervention(year=1970, capital_investment_generation=0.06),
    state_sampler=sample_initial_states,
    propensity_scorer=0.25,
    outcome_extractor=lambda run: run[2000].population,
    covariate_builder=lambda run: run.initial_state.values())


###########################
# Treatment Bias Experiment
###########################
@parameter(name="treatment_bias", default=0.2, values=[0.1, 0.2, 0.5, 0.9],
           description="How much to bias treatment assignment")
@parameter(name="treatment_percentage", default=0.2, values=[0.05, 0.2, 0.5, 0.9],
           description="What fraction of the units should be treated.")
def biased_treatment_propensity(control_runs, treatment_bias, treatment_percentage):
    """Compute propensity scores as a function of the bias."""
    # Sort runs by initial quantity of natural resources
    sorted_idxs = sorted(range(len(control_runs)),
                         key=lambda idx: control_runs[idx].initial_state.natural_resources,
                         reverse=True)

    # More likely to treat units with top initial quantity of natural resources
    top_resource_idxs = sorted_idxs[int(treatment_percentage * len(sorted_idxs)):]

    propensities = (1. - treatment_bias) * np.ones(len(control_runs))
    propensities[top_resource_idxs] = treatment_bias

    return propensities


# pylint: disable-msg=invalid-name
#: Observational experiment with confounding for world2.
BiasedTreatment = DynamicsExperiment(
    name="world2_biased_treatment",
    description="Experiment with treatment bias on World 2.",
    simulator=world2,
    simulator_config=world2.Config(),
    intervention=world2.Intervention(year=1970, capital_investment_generation=0.06),
    state_sampler=sample_initial_states,
    propensity_scorer=biased_treatment_propensity,
    outcome_extractor=lambda run: run[2000].population,
    covariate_builder=lambda run: run.initial_state.values())


######################
# Mediation Experiment
######################
@parameter(name="birth_rate_intervention", default=0.02, values=[0.039, 0.02, 0.01],
           description="Decrease in birth rate in the intervention year.")
def mediation_intervention(birth_rate_intervention):
    """Return the treatment config for the mediation experiment.

    Provides an example showing how the framework supports parameterized
    configurations.
    """
    base_birth_rate = world2.Config().birth_rate
    assert np.less_equal(birth_rate_intervention, base_birth_rate)
    return world2.Intervention(year=1970, birth_rate=birth_rate_intervention)


def mediation_propensity_scores(intervention, control_runs):
    """Probability of treating each unit.

    Units with the largest populations are more likely to be treated.
    """
    populations = [run[intervention.year].population for run in control_runs]
    upper_quantile = np.quantile(populations, 0.9)
    propensities = 0.05 * np.ones(len(control_runs))
    propensities[populations > upper_quantile] = 0.9
    return propensities


@parameter(name="mediation_year", default=1980, values=[1980, 2000, 2030, 2030],
           description="Year to choose the mediation states.")
@parameter(name="num_mediators", default=3, values=[0, 1, 2, 3, 4],
           description="How many mediating states to choose.")
def mediation_covariates(run, intervention, mediation_year, num_mediators):
    """Return covariates including mediators."""
    confounders = run[intervention.year].values()
    mediators = run[mediation_year].values()
    # Use only a subset of the mediators to avoid blocking all paths
    return np.concatenate([confounders, mediators[:num_mediators]])


# pylint: disable-msg=invalid-name
#: Observational experiment with mediation for world2.
Mediation = DynamicsExperiment(
    name="world2_mediation",
    description="Create mediation by adding covariates after intervention.",
    simulator=world2,
    simulator_config=world2.Config(),
    intervention=mediation_intervention,
    state_sampler=sample_initial_states,
    propensity_scorer=mediation_propensity_scores,
    outcome_extractor=lambda run: run[2030].quality_of_life,
    covariate_builder=mediation_covariates)

if __name__ == "__main__":
    RCT.run(num_samples=10)



"""Simple integration and determinism tests for all simulators."""
import numpy as np

import whynot as wn


def check_shapes(dataset, num_samples):
    """Ensure datast has the right size."""
    assert len(dataset.covariates.shape) == 2 and dataset.covariates.shape[0] == num_samples
    assert dataset.treatments.shape == (num_samples,)
    assert dataset.outcomes.shape == (num_samples,)
    assert dataset.true_effects.shape == (num_samples,)

def integration_test(simulator, num_samples=15):
    """Check if the simulator generates data that is well-formed and can be used for inference."""
    for experiment in simulator.get_experiments():
        params = experiment.get_parameters()
        print(params)
        dataset = experiment.run(num_samples=num_samples, seed=1234, **params.sample())
        check_shapes(dataset, num_samples)

        # Sanity check to make sure data is well-formed.
        _ = wn.algorithms.ols.estimate_treatment_effect(
            dataset.covariates, dataset.treatments, dataset.outcomes)

def determinism_test(simulator, num_samples=15):
    """Check to make sure all experiments are deterministic."""
    for experiment in simulator.get_experiments():
        dataset1 = experiment.run(num_samples=num_samples, seed=1234)
        dataset2 = experiment.run(num_samples=num_samples, seed=1234)

        assert np.allclose(dataset1.covariates, dataset2.covariates)
        assert np.allclose(dataset1.treatments, dataset2.treatments)
        assert np.allclose(dataset1.outcomes, dataset2.outcomes)
        assert np.allclose(dataset1.true_effects, dataset2.true_effects)

# pylint:disable-msg=missing-docstring
def test_civil_violence_integration():
    integration_test(wn.civil_violence, num_samples=5)

def test_civil_violence_determinism():
    determinism_test(wn.civil_violence, num_samples=5)

def test_dice_integration():
    integration_test(wn.dice)

def test_dice_determinism():
    determinism_test(wn.dice)

def test_hiv_integration():
    integration_test(wn.hiv)

def test_hiv_determinism():
    determinism_test(wn.hiv)

def test_lalonde_integration():
    # 445 units in the LaLonde dataset
    integration_test(wn.lalonde, num_samples=445)

def test_lalonde_determinism():
    determinism_test(wn.lalonde)

def test_lotka_volterra_integration():
    integration_test(wn.lotka_volterra)

def test_lotka_volterra_determinism():
    determinism_test(wn.lotka_volterra)

def test_opioid_integration():
    integration_test(wn.opioid)

def test_opioid_determinism():
    determinism_test(wn.opioid)

def test_schelling_integration():
    integration_test(wn.schelling, num_samples=5)

def test_schelling_determinism():
    determinism_test(wn.schelling, num_samples=5)

def test_simpact_integration():
    integration_test(wn.simpact_hiv)

def test_simpact_determinism():
    determinism_test(wn.simpact_hiv)

def test_world2_integration():
    integration_test(wn.world2)

def test_world2_determinism():
    determinism_test(wn.world2)

def test_world3_integration():
    integration_test(wn.world3, num_samples=25)

def test_world3_determinism():
    determinism_test(wn.world3, num_samples=25)
